(* Routines to convert between the SVG and smoke formats                    *)

open List
open Path

exception ConversionError of string

let colour rgba = 
  { Smoke.red = rgba.Svgml.red;
    Smoke.green = rgba.Svgml.green;
    Smoke.blue = rgba.Svgml.blue;
    Smoke.alpha = rgba.Svgml.alpha }

let white = 
  { Smoke.red = 1.0; Smoke.green = 1.0;
    Smoke.blue = 1.0; Smoke.alpha = 1.0 }

let black = 
  { Smoke.red = 0.0; Smoke.green = 0.0;
    Smoke.blue = 0.0; Smoke.alpha = 1.0 }


let vector xy = xy.x, xy.y

let smoke_style scene style_list = 
  (* Set the defaults. currently ignoring CSS etc! *)
  (* Nb. Smoke line widths are currently half SVG's *)
  let fill = ref (Smoke.FlatFill black)
  and stroke = ref None
  and line_width = ref 1.0
  and join_style = ref Smoke.JoinMitred
  and endcap_style = ref Smoke.EndcapButt
  in
  let process_style = function
    | Svgml.Fill rgba ->
	fill := Smoke.FlatFill (colour rgba)
    | Svgml.Stroke rgba ->
	stroke := Some (Smoke.FlatFill (colour rgba))
    | Svgml.StrokeWidth width ->
	line_width := width
  in iter process_style style_list;
  let fill_index = Smoke.add_fill_type scene !fill
  and stroke_index = match !stroke with
  | None -> None
  | Some x -> Some (Smoke.add_fill_type scene x) in
  let line_style_index = match stroke_index with
  | None -> []
  | Some x -> [ Smoke.add_line_style scene
		  { Smoke.line_fill = x;
		    Smoke.line_width = !line_width /. 2.0; (* See above *)  
		    Smoke.line_join_style = !join_style;
		    Smoke.line_endcap_style = !endcap_style } ]
  in fill_index, line_style_index

let path_to_contourObject fills styles path = 
  (* In SVG a path consists of one or more subpaths, analagous to smoke     *)
  (* contours. Contours are delimited by either an explicit ClosePath (in   *)
  (* which case they are like Closed smoke contours) or implicitly delimited*)
  (* by a MoveTo element (in which case they are like Open smoke contours)  *)


  let subpaths = 

    let rec path_to_subpaths subpaths current path = match path with
    | (ClosePath _ as x)::rest ->
	path_to_subpaths (subpaths@ [(current@[x])]) [] rest
    | (MoveTo (_,_) as x)::rest ->
	path_to_subpaths (subpaths@ [(current@[x])]) [] rest
    | x::rest ->
	path_to_subpaths subpaths (current@[x]) rest
    | [] -> 
	if current = [] then subpaths else (subpaths@[current])
    
    in match path with
    | [] -> []
    | (x::xs) -> path_to_subpaths [] [x] xs

  in
  let subpath_to_contour subpath = 
    (* In SVG a subpath is Closed iff it ends in a ClosePath element. *)
    (* In Smoke a subpath is explicitly labelled as Closed.           *)
    (* In SVG all subpaths must start with a MoveTo.                  *)
    (* In Smoke an Open contour has a start vector while a Closed     *)
    (*   contour is placed and the ending coordinate of the last      *)
    (*   segment.                                                     *)
    let start_vec = match subpath with 
    | (MoveTo (_, vec) :: _) -> vec
    | _ -> raise (ConversionError "SVG paths must begin with a MoveTo") in
    let rest = tl subpath in

    let is_closed = ref false in

    (* Smoke currently suffers numerical instability with 0 length    *)
    (* lines. This is triggered by tiger.svg's use of both "manual"   *)
    (* (with a bezier endpoint) and "automatic" (with a ClosePath)    *)
    (* methods of path closure. This function erases the terminating  *)
    (* redundant, 0 length line segments.                             *)
    let rec remove_redundant_line = function
      (* nothing to do in trivial cases *)
      |	[] -> []
      |	[ x ] -> [ x ]

      (* cases where the last element is redundant... *)
      |	[ Smoke.Line a; Smoke.Line b ] when a = b   -> [ Smoke.Line a ]
      |	[ Smoke.Bezier (c1, c2, a); Smoke.Line b ] when a = b -> 
	  [ Smoke.Bezier (c1, c2, a) ]
      (* ... and when it isn't (the previous 2 failed to match) *)
      |	[ x; y ] -> [ x; y ]

      (* not found the end of the list yet - keep looking *)
      |	x :: xs -> x :: (remove_redundant_line xs)


    in let rec elements_to_segments pen_pos elements =
      let relative xy = { x = xy.x +. pen_pos.x;
			  y = xy.y +. pen_pos.y }
      in match elements with
      |	[] -> []
      |	LineTo (Absolute, xy) :: rest ->
	  Smoke.Line (vector xy) :: (elements_to_segments xy rest)
      |	LineTo (Relative, xy) :: rest ->
	  Smoke.Line (vector (relative xy)) ::
	  (elements_to_segments (relative xy) rest)
      |	[ ClosePath _ ] ->
	  is_closed := true;
	  [ Smoke.Line (vector start_vec) ]
      |	ClosePath _ :: _ ->
	  raise (ConversionError "Subpaths should end with ClosePaths!")
      |	HorizLineTo (Absolute, x) :: rest ->
	  Smoke.Line (vector ({x = x; y = pen_pos.y})) ::
	  (elements_to_segments ({x = x; y = pen_pos.y})) rest
      |	HorizLineTo (Relative, x) :: rest ->
	  Smoke.Line (vector ({ x = pen_pos.x+.x; y = pen_pos.y})) ::
	  (elements_to_segments ({x = pen_pos.x; y = pen_pos.y})) rest
      |	VertLineTo (Absolute, y) :: rest ->
	  Smoke.Line (vector ({x = pen_pos.x; y = y})) ::
	  (elements_to_segments ({x = pen_pos.x; y = y})) rest
      |	VertLineTo (Relative, y) :: rest ->
	  Smoke.Line (vector ({ x = pen_pos.x; y = pen_pos.y+.y})) ::
	  (elements_to_segments ({x = pen_pos.x; y = pen_pos.y+.y})) rest
      |	CurveTo (Absolute, (a, b, c)) :: rest ->
	  Smoke.Bezier (vector a, vector b, vector c) ::
	  (elements_to_segments c rest)
      |	CurveTo (Relative, (a, b, c)) :: rest ->
	  let a', b', c' = relative a, relative b, relative c in
	  Smoke.Bezier (vector a', vector b', vector c') ::
	  (elements_to_segments c' rest)
      
      |	MoveTo (_, _) :: _ ->
	  raise (ConversionError "Subpaths dont have MoveTos!")

      |	ShorthandCurveTo (_, _) :: _ ->
	  raise (ConversionError "ShorthandCurveTo path-data not supported")
      |	QuadCurveTo (_, _) :: _->
	  raise (ConversionError "QuadCurveTo path-data not supported")
      |	ShorthandQuadCurveTo (_, _) :: _ ->
	  raise (ConversionError "ShorthandQuadCurveTo path-data not supported")
      |	Arc (_, _) :: _->
	  raise (ConversionError "Arc path-data not supported")
    in
    let segments = remove_redundant_line
	(elements_to_segments start_vec rest) in
    if !is_closed then (Smoke.Closed segments)
    else (Smoke.Open (vector start_vec, segments))

  in
  let all_contours = map subpath_to_contour subpaths
  in Smoke.ContourObject( fills, styles, all_contours)

let rec add_obj_to_scene scene style fill = function
  | Svgml.Path pt ->
      let contour_object = path_to_contourObject [ fill ] style  
	  pt.Svgml.path_data 
      in Smoke.add_object scene contour_object
  | Svgml.Group gt ->
      (* do we inherit styles in SVG? *)
      let fill, style = smoke_style scene gt.Svgml.group_style
      in iter (fun x -> add_obj_to_scene scene style fill x)
	gt.Svgml.group_objects
  | Svgml.Svg gt ->
      (* do we inherit styles in SVG? *)
      let fill, style = smoke_style scene gt.Svgml.group_style
      in iter (fun x -> add_obj_to_scene scene style fill x)
	gt.Svgml.group_objects

let generate_smoke_scene svg = 
  let scene = Smoke.blank_scene() in
  let default_fill, default_linestyle = smoke_style scene [] in
  add_obj_to_scene scene default_linestyle default_fill svg;
  scene
  
