(* Lexer for SVG Path data *)

{
  open Path
  open Pathparse        (* The type token is defined in pathparse.mli *)
  exception Eof

  (* A function to return the current file position *)
  let p lexbuf = lexbuf.Lexing.lex_curr_pos
}
  
let digit = ['0'-'9']
let plusminus = ['-' '+']

(* FIXME -- is this right for IEEE floats? *)
let exponent = ( 'e' | 'E' ) plusminus? digit+
let floating = plusminus? digit+ ('.' digit+)? exponent?

(* Commas appear to be used as separators and may be dropped when not
   ambiguous... I think we can skip them as long as they are still allowed
   to delimit tokens (like whitespace also does) *)
let whitespace = [ ' ' '\t' '\n' ',' ]

rule token = parse

(* Path data ignores whitespace *)
| whitespace       { token lexbuf }

| floating         { NUMBER (float_of_string (Lexing.lexeme lexbuf) ) }

| "M"              { MOVE (Absolute) }
| "m"              { MOVE (Relative) }
| "Z"              { CLOSE (Absolute) }
| "z"              { CLOSE (Relative) }
| "L"              { LINE (Absolute) }
| "l"              { LINE (Relative) }
| "H"              { HORIZLINE (Absolute) }    (* Shortcut horizontal lines *)
| "h"              { HORIZLINE (Relative) }
| "V"              { VERTLINE (Absolute) }     (* Shortcut vertical lines *)
| "v"              { VERTLINE (Relative) }
| "C"              { CURVE (Absolute) }
| "c"              { CURVE (Relative) }

(* From the spec:
   The first control point is assumed to be the reflection of the second 
   control point on the previous command relative to the current point.
   (If there is no previous command or if the previous command was not an 
   C, c, S or s, assume the first control point is coincident with the 
   current point.)
*)
| "S"              { SHORTHANDCURVE (Absolute) } 
| "s"              { SHORTHANDCURVE (Relative) }

| "Q"              { QUADCURVE (Absolute) }
| "q"              { QUADCURVE (Relative) }
| "T"              { SHORTHANDQUADCURVE (Absolute) }
| "t"              { SHORTHANDQUADCURVE (Relative) }

| "A"              { ARC (Absolute) }
| "a"              { ARC (Relative) }

| eof              { EOF }
