(* Basic SVG types defined in http://www.w3.org/TR/SVG/types.html            *)

(* ... the range for a <integer> encompasses (at a minimum)                  *)
(* -2147483648 to 2147483647.                                                *)
type integer = int

type number = float

(* http://www.w3.org/TR/SVG/coords.html#Units                                *)
(* The list of unit identifiers in SVG matches the list of unit identifiers  *)
(* in CSS: em, ex, px, pt, pc, cm, mm, in and percentages.                   *)
(* (a missing unit identifier means "user units"                             *)
type unit_type = Em | Ex | Px | Pt | Pc | Cm | Mm | In | Percent | User
 
type length = number * unit_type

type coord = length * length

(* For properties defined in [CSS2], an angle unit identifier must be        *)
(* provided. For SVG-specific attributes and properties, the angle unit      *)
(* identifier is optional. If not provided, the angle value is assumed to be *)
(* in degrees.                                                               *)

type angle_unit_type = Degrees | Grads | Radians

type angle = number * angle_unit_type

type byte = int

type colour = {red: byte; green: byte; blue: byte}

type uri = string

(* FIXME: What shall we do with <percentage> <transform-list> <paint>?       *)

(* Time-related types used presumably for animation?                         *)
type frequency_type = Hertz | KiloHertz

type frequency = number * frequency_type

type time_type = Seconds | MilliSeconds

type time = number * time_type

(* ------------------------------------------------------------------------- *)
(* Pretty-printing and parsing functions                                     *)

module type PRINT = sig
  val integer:         integer         -> string
  val number:          number          -> string
  val unit_type:       unit_type       -> string
  val length:          length          -> string
  val coord:           coord           -> string
  val angle_unit_type: angle_unit_type -> string
  val angle:           angle           -> string
  val colour:          colour          -> string
  val uri:             uri             -> string
  val frequency_type:  frequency_type  -> string
  val frequency:       frequency       -> string
  val time_type:       time_type       -> string
  val time:            time            -> string
end

module Print: PRINT = struct
  
  let integer = string_of_int

  let number = string_of_float

  let unit_type = function
    | Em -> "Em" | Ex -> "Ex" | Px -> "Px" | Pt -> "Pt" | Cm -> "Cm"
    | Mm -> "Mm" | In -> "In" | Percent -> "%" | Pc -> "Pc" | User -> ""

  let length (n, u) = (number n) ^ (unit_type u)

  let coord (x, y) = (length x) ^ " x " ^ (length y)

  let angle_unit_type = function
    | Degrees -> "Degrees" | Grads -> "Grads" | Radians -> "Radians"

  let angle (n, u) = (number n) ^ " " ^ (angle_unit_type u)

  let byte b = 
    let high_nibble = b / 16
    and low_nibble = b mod 16
    and char_of_nibble x = 
      if x <= 9
      then char_of_int (int_of_char '0' + x)
      else char_of_int (int_of_char 'a' + x - 10)
    in
    let high_char = char_of_nibble high_nibble
    and low_char  = char_of_nibble low_nibble
    and result = String.create 2
    in
    String.set result 0 high_char;
    String.set result 1 low_char;
    result

  let colour c = "#" ^ (byte c.red) ^ (byte c.green) ^ (byte c.blue)
  let uri x = x

  let frequency_type = function Hertz -> "Hz" | KiloHertz -> "kHz"

  let frequency (n, u) = (number n) ^ (frequency_type u)

  let time_type = function Seconds -> "s" | MilliSeconds -> "ms"

  let time (n, u) = (number n) ^ (time_type u)
end

exception InvalidSVGValue of string (* typename *) * string (* value *)
exception UnimplementedSVGValue of string (* typename *) * string (* value *)

module type PARSE = sig
  val integer:         string -> integer
  val number:          string -> number
  val unit_type:       string -> unit_type
  val length:          string -> length
  val coord:           string -> coord
  val angle_unit_type: string -> angle_unit_type
  val angle:           string -> angle
  val colour:          string -> colour
  val uri:             string -> uri
  val frequency_type:  string -> frequency_type
  val frequency:       string -> frequency
  val time_type:       string -> time_type
  val time:            string -> time
end

module Parse: PARSE = struct

  (* Util function: table lookup, case insensitive *)
  let rec lookup table key = match table with
  | [] -> None
  | ((x,y)::xs) -> if (String.lowercase x) = (String.lowercase key) then Some y
  else lookup xs key

  let unit_type_table = [ "Em", Em; "Ex", Ex; "Px", Px; "Pt", Pt; "Pc", Pc;
			  "Cm", Cm; "Mm", Mm; "In", In; "%", Percent; "", User]

  let integer = int_of_string
  let number = float_of_string
  let unit_type x = raise (UnimplementedSVGValue ("unit", x))
  let length x = raise (UnimplementedSVGValue ("length", x))
  let coord x = raise (UnimplementedSVGValue ("coord", x))
  let angle_unit_type x = raise (UnimplementedSVGValue ("angle unit", x))
  let angle x = raise (UnimplementedSVGValue ("angle", x))
  let colour x = raise (UnimplementedSVGValue ("colour", x))
  let uri x = raise (UnimplementedSVGValue ("uri", x))
  let frequency_type x = raise (UnimplementedSVGValue ("frequency type", x))
  let frequency x = raise (UnimplementedSVGValue ("frequency", x))
  let time_type x = raise (UnimplementedSVGValue ("time type", x))
  let time x = raise (UnimplementedSVGValue ("time", x))


end
