
open Pxp_yacc 
open Pxp_types
open Pxp_document

open Svgml

open List

exception ParseFailure of string

(* ------------------------------------------------------------------------ *)
(* Interface with the PXP XML parser.                                       *)

class display_xml_warnings = 
  object
    method warn w = print_endline("WARNING: " ^ w)
  end

(* XML filename -> DTD filename -> XML document list                        *)
(* Note that it's running in validating mode, so it's a bit slow.           *)
let parse filename dtd_filename = 
  try
    let dtd = parse_dtd_entity default_config (from_file dtd_filename) in
    let config = { default_config with warner = new display_xml_warnings } in
    parse_content_entity config (from_file filename) dtd default_spec
  with
    e -> (print_endline (Pxp_types.string_of_exn e); raise e)

(* ------------------------------------------------------------------------ *)
(* Utility functions for dealing with some of the XML data.                 *)

(* Returns true if the XML document node is an element with name, name      *)
let node_name_is node name = 
  match (node # node_type) with
  | T_element x -> x = name
  | _ -> false

(* Reads in a ;-delimited list of key:value pairs into a string*string Hash *)
(* Assumes that style expressions are always in this form! Does the spec    *)
(* reference all of CSS?                                                    *)
let style_to_hash text = 
  let properties = Hashtbl.create 10 in
  let items = Str.split (Str.regexp "[ \t\n]*;[ \t\n]*") text in
  let add_property x = 
    let keyval = Str.split (Str.regexp "[ \t\n]*:[ \t\n]*") x in
    Hashtbl.add properties (hd keyval) (hd (tl keyval))
  in
  List.iter add_property items;
  properties


(* Reads a 24-bit colour string into a colour value. How does SVG define    *)
(* alpha values?                                                            *)
let rgbstring_to_colour x =
  (* Assumes colour is prefixed with a # symbol *)
  let bits = Str.split (Str.regexp "#") x in
  let colour = match bits with
  | [ c ] -> c
  | [ _; c ] -> c
  in
  
  let rr = String.sub colour 0 2
  and gg = String.sub colour 2 2
  and bb = String.sub colour 4 2
 
  and int_from_hex x = 
    let h = String.get x 0
    and l = String.get x 1 in
    let int_from_nibble x =
      if (Pervasives.compare x '9' <= 0)
      then int_of_char x - int_of_char '0'
      else int_of_char x - int_of_char 'a' + 10 in
    (int_from_nibble h)*16 + (int_from_nibble l)
  and col_int_to_float x = (float_of_int x) /. 255.0
    
  in
  { red = col_int_to_float (int_from_hex rr);
    green = col_int_to_float (int_from_hex gg);
    blue = col_int_to_float (int_from_hex bb);
    alpha = 1.0 }

(* Converts the string*string style Hash into a list of style_option types  *)
let style_hash_to_style_list hash = 
  let make_style k v = match k with
  | "fill" -> Fill (rgbstring_to_colour v)
  | "stroke" -> Stroke (rgbstring_to_colour v)
  | "stroke-width" -> StrokeWidth (float_of_string v)
  | x -> raise (ParseFailure ("Unknown style \"" ^ x ^ "\""))
  in 
  Hashtbl.fold (fun k v rst -> (make_style k v) :: rst) hash [] 

(* Wrapper function which returns styles attached to any supplied XML node  *)
let read_style_list node = 
  match (node # attribute "style") with
  | Value styles ->
      let table = style_to_hash styles in
      style_hash_to_style_list table
  | _ -> [] (* no styles *)

(* ------------------------------------------------------------------------ *)
(* Recursive-descent parse of the main SVG elements                         *)

let rec read_group_type node = 
  let styles = read_style_list node in
  let children = node # sub_nodes in
  { group_style = styles;
    group_objects = map read_obj children }

and read_g node = 
  assert (node_name_is node "g");
  Group (read_group_type node)

and read_svg node = 
  assert (node_name_is node "svg");
  Svg (read_group_type node)

and read_path node = 
  assert (node_name_is node "path");

  let read_path_list node = 
    match (node # attribute "d") with
    | Value pathdata_string ->
	let lexbuf = Lexing.from_string pathdata_string in
	let path_list = Pathparse.path Pathlex.token lexbuf in
	path_list
    | _ -> [] (* empty path *)
  in
  Path { path_data = read_path_list node }


and read_obj node = 
  let by_name = function
    | "g" -> read_g node
    | "svg" -> read_svg node
    | "path" -> read_path node
    | x -> raise (ParseFailure ("No code to handle node named \"" ^ x ^ "\""))

  in match (node # node_type) with
  | T_element x -> by_name x
  | _ -> raise (ParseFailure "Unexpected SVG Node (not an element)")

  
(* ------------------------------------------------------------------------ *)

let test_parse n = 
  let toplevel = read_obj n in
  print_string (pp_obj toplevel)

(* ------------------------------------------------------------------------ *)

(* Debug routine which outputs all the XML as text.                         *)
let rec print_structure n = 
  let rec print_level level n =
    let rec maken char n = 
      match n with
	0 -> ""
      |	n -> char ^ (maken char (n-1))
    in
    let ntype = n # node_type in
    match ntype with 
      T_element name -> 
	print_endline((maken " " level) ^ "Element of type: "^name);
        let children = n # sub_nodes in
        List.iter (print_level (level+1)) children
    | T_data -> 
	print_endline ( (maken " " level) ^ "Data = " ^ (n #data))
    |  _ -> 
        (* hopefully nothing *) assert false
  in print_level 0 n

(* ------------------------------------------------------------------------ *)
(* Entrypoint of the code                                                   *)

let parse_svg_file filename = 
  let documents = parse filename "svg10.dtd" in
  read_obj (documents)

(*
let main() =
  let files = ref [] in
  Arg.parse [] (fun x -> files := x :: !files) "
usage: parse file ...

";
  files := List.rev !files;
  let documents = List.map (fun fn -> parse fn "svg10.dtd")!files in
(*  List.iter print_structure documents *)
  List.iter test_parse documents 


(* This top level binding causes the program to be executed                 *)
let _ = main() 


*)
