/** file: autoptr.h
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#ifndef AUTOPTR
#define AUTOPTR



/** Generic pointer type which can be responsible for destructing the object
 *  pointed to.
 */
template <class T>
class AutoPtr
{
 private:
  /** Internal representation of the pointer. */
  mutable T* ptr;

 public:
  /** The type of the element pointed to by this autoptr. */
  typedef T element_type;
  
  /** Default constructor. */
  AutoPtr(T* p = 0);
  
  /** Copy contructing an AutoPtr from another AutoPtr causes the old one
   *  to be released from destruction duty. */
  AutoPtr(const AutoPtr &a);

  template<class T1>
    AutoPtr(const AutoPtr<T1> &a);

  /** Implicit cast to a pointer. */
  T *
    operator()(T *);

  /** Assignment of one AutoPtr to another destructs the object currently
   *  pointed to and takes over from the other pointer. */
  AutoPtr&
    operator=(AutoPtr &a);

  /** Assignment to a pointer with a different element type. */  
  template <class T1> 
   AutoPtr &
    operator=(AutoPtr<T1> &a);

  /** Destruction of an AutoPtr also destroys any data pointed to. */
  ~AutoPtr();
  
  /** Dereference. */
  T& operator*() const;
  
  /** Dereference. */
  T* operator->() const;

  /** Implicit cast to a pointer. */
  operator T *();
  
  /** Release the current AutoPtr from destruction duty. */
  T* release() const;

  /** Destruct any object pointed to and Reset this AutoPtr to NULL. */
  void reset(T * p = 0);
};



#include "autoptr.cc"



#endif /* AUTOPTR */
