/** fill_types.cc
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "fill_types.h"

#include <GL/gl.h>



// FillType

FillType::~FillType()
{
}



// FlatFill

FlatFill::FlatFill(const Colour c)
  : colour(c)
{
}

bool
FlatFill::isOpaque() const
{
  return (colour.getUInt()&0xff000000)==0xff000000;
}

FillType *
FlatFill::clone() const
{
  return new FlatFill(colour);
}

void
FlatFill::set(const ViewPort &viewport) const
{
  // FIXME: This is faster than the alternatives but will only work in one
  // endianess
  glColor4ubv((GLubyte *)&colour);
}

void
FlatFill::unset() const
{
  glColor4f(1, 1, 1, 1);
}

void
FlatFill::save(std::ostream &out,
	       const Store *store) const
{
  out << "FlatFill\n"
      << colour.getRed() << " "
      << colour.getGreen() << " "
      << colour.getBlue() << " "
      << colour.getAlpha();
}



// Gradient Fill

GradientFill::GradientFill(const Colour c0,
			   const Colour c1,
			   const Vec a,
			   const Vec b)
  : colour0(c0),
    colour1(c1),
    r0(a),
    r1(b),
    texture(256)
{
  // Create a texture map containing the gradient
  for (int i=0; i<256; i++)
    {
      float t=i/255.0;
      // TODO: Xara X supports a kind of non-linear 0->1 mapping of parameters
      // like `t'. Why don't we? :)
      Colour c(colour0.getRed()*(1-t) + colour1.getRed()*t,
	       colour0.getGreen()*(1-t) + colour1.getGreen()*t,
	       colour0.getBlue()*(1-t) + colour1.getBlue()*t,
	       colour0.getAlpha()*(1-t) + colour1.getAlpha()*t);

      texture[i] = c.getUInt();
    }

  glGenTextures(1, &texture_id);
  glBindTexture(GL_TEXTURE_1D, texture_id);
  glTexParameteri(GL_TEXTURE_1D,
		  GL_TEXTURE_MIN_FILTER,
		  GL_NEAREST);
  glTexParameteri(GL_TEXTURE_1D,
		  GL_TEXTURE_MAG_FILTER,
		  GL_NEAREST);
  glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexImage1D(GL_TEXTURE_1D,
	       0,
	       GL_RGBA,
	       256,
	       0,
	       GL_RGBA,
	       GL_UNSIGNED_BYTE,
	       (GLubyte *)&(texture.front()));
}

bool
GradientFill::isOpaque() const
{
  return
    (colour0.getUInt()&0xff000000)==0xff000000 &&
    (colour1.getUInt()&0xff000000)==0xff000000;
}

FillType *
GradientFill::clone() const
{
  return new GradientFill(colour0, colour1, r0, r1);
}

void
GradientFill::set(const ViewPort &viewport) const
{
  // Calculate the scaled normal to the "plane" s=0 and the translation factor
  // `q'
  Vec n=r1-r0;
  n/=dot(n, n);
  Num q=-dot(r0, n);

  // Generate the `s' coordinate using the calculated object plane and
  // place the transformation (in the direction of the normal) in the texture
  // matrix
  TinyVector<GLfloat, 4> p(n(0), n(1), 0, 0);
  glMatrixMode(GL_TEXTURE);
  glPushMatrix();
  // FIXME: Can't we just do a glLoadIdentity() here instead of pushing and
  // popping? Would we ever need to use a previously setup texture matrix?
  glTranslatef(q, 0, 0);
  glMatrixMode(GL_MODELVIEW);

  // The current model view matrix allows us to affine transform the _fill
  // type_ so we set it back to the identity
  //glPushMatrix();
  //glLoadIdentity();
  //glTranslatef(0, -.5, 0);
  //glRotatef(180, 0, 0, 1);

  // Bind our texture and initialize texture generation for the `s' coordinate
  glBindTexture(GL_TEXTURE_1D, texture_id);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  // We use texture coordinate generation in eye space which is in canvas
  // coordinates
  glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
  glTexGenfv(GL_S, GL_EYE_PLANE, (GLfloat *)&p);

  // Enable the appropriate texturing
  glEnable(GL_TEXTURE_1D);
  glEnable(GL_TEXTURE_GEN_S);

  // Recover the original model matrix which will now be applied to the object
  // to be rendered
  //glPopMatrix();

  glColor4f(1, 1, 1, 1);
}

void
GradientFill::unset() const
{
  // Get the original texture matrix back
  glMatrixMode(GL_TEXTURE);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);

  // Turn off 1D texturing and texture coordinate generation
  glDisable(GL_TEXTURE_1D);
  glDisable(GL_TEXTURE_GEN_S);
}

void
GradientFill::save(std::ostream &out,
		   const Store *store) const
{
  out << "GradientFill\n"
      << colour0.getRed() << " "
      << colour0.getGreen() << " "
      << colour0.getBlue() << " "
      << colour0.getAlpha() << " "
      << r0(0) << " " << r0(1) << "\n"
      << colour1.getRed() << " "
      << colour1.getGreen() << " "
      << colour1.getBlue() << " "
      << colour1.getAlpha() << "\n"
      << r1(0) << " " << r1(1);
}
