/** fill_types.h
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

class FillType;
class FlatFill;
class GradientFill;

#ifndef FILL_TYPES_H
#define FILL_TYPES_H

class Store;

#include "colour.h"
#include "vecmat.h"
#include "viewport.h"

#include <vector>
#include <GL/gl.h>



/** Scalar field of colours with which to fill objects.
 *
 *  TODO: Move OpenGL state construction out of the contructors of the fill
 *  types and into initGL member functions so that vector graphics scenes can
 *  be used without requiring an OpenGL context. A finishGL call may be
 *  necessary to remove the state if the scene is to be used in multiple GL
 *  contexts.
 */
class FillType
{
 public:
  virtual
    ~FillType();

  /** Allocate and initialise an identical copy of this fill type. */
  virtual
    FillType *
    clone() const = 0;

  /** Detect an entirely opaque fill type.
   *
   *  Costly blending and depth buffering operations can be optimised out for
   *  entirely opaque fill types.
   */
  virtual
    bool
    isOpaque() const = 0;

  /** Initialise OpenGL for rendering using this fill type. */
  virtual
    void
    set(const ViewPort &viewport) const = 0;

  /** Remove any state changes. */
  virtual
    void
    unset() const = 0;

  virtual
    void
    save(std::ostream &out,
	 const Store *store) const = 0;
};



/** Constant colour. */
class FlatFill : public FillType
{
 private:
  const Colour colour;

 public:
  FlatFill(const Colour c);

  bool
    isOpaque() const;

  FillType *
    clone() const;

  void
    set(const ViewPort &viewport) const;
  
  void
    unset() const;

  void
    save(std::ostream &out,
	 const Store *store) const;
};



/** Clamped linear gradient in the direction of a line.
 */
class GradientFill : public FillType
{
 private:
  const Colour colour0, colour1;
  const Vec r0, r1;
  std::vector<unsigned int> texture;
  GLuint texture_id;

 public:
  GradientFill(const Colour c0,
	       const Colour c1,
	       const Vec a,
	       const Vec b);

  bool
    isOpaque() const;

  FillType *
    clone() const;

  void
    set(const ViewPort &viewport) const;
  
  void
    unset() const;

  void
    save(std::ostream &out,
	 const Store *store) const;
};



#endif /* FILL_TYPES_H */
