/** file: geometry.cc
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "geometry.h"

#include "tesselate.h"

#include <cmath>
#include <iostream>


// TesselationCache

TesselationCache::TesselationCache()
  : scale(2),
    central_deviation(0)
{
}

const TesselationCache::TesselationCache2 *
TesselationCache::renderEntry(Num d) const
{
  if (tesselations.empty())
    return NULL;

  // Calculate the entry in the initial deque from the x components
  int entry=0;
  if (central_deviation != 0 && d != 0)
    entry = (int)(floor(log(d/central_deviation)/log(scale)));

  // If the entry is out of bounds then return, failing to render
  if (entry<0 || entry>=(int)tesselations.size())
    return NULL;

  return &tesselations[entry];
}

bool
TesselationCache::render(GLenum winding_rule,
			 Vec delta) const
{
  // Try to find an exact tesselation if any are present
  if (exact_tesselations.render(winding_rule, delta(1)))
    return true;

  // If an exact tesselation could not be found then try to find an approximate
  // one
  const TesselationCache2 *t=renderEntry(delta(0));
  if (!t) return false;
  return t->render(winding_rule, delta(1));
}

bool
TesselationCache::render(const StrokeStyle &style,
			 Vec delta) const
{
  // Try to find an exact tesselation if any are present
  if (exact_tesselations.render(style, delta(1)))
    return true;

  // If an exact tesselation could not be found then try to find an approximate
  // one
  const TesselationCache2 *t=renderEntry(delta(0));
  if (!t) return false;
  return t->render(style, delta(1));
}

int
TesselationCache::getIndex(Num d) const
{
  return int(floor(log(d/central_deviation)/log(scale)));
}

TesselationCache::Range
TesselationCache::insertEntry(Num dev,
			      Num del)
{
  // If there are no entries then we will create them with the first entry
  // representing the given cached deviation
  if (tesselations.empty()) central_deviation = dev;

  int entry1=getIndex(dev), entry2=getIndex(del);
  while (entry1<0 || entry2<0)
    {
      entry1++;
      entry2++;
      tesselations.push_front(TesselationCache2());
      central_deviation /= scale;
    }
  while ((unsigned int)entry1>=tesselations.size() ||
	 (unsigned int)entry2>=tesselations.size())
    tesselations.push_back(TesselationCache2());
  return Range(tesselations.begin() + entry1,
	       ++tesselations.begin() + entry2);
}

void
TesselationCache::insert(GLenum winding_rule,
			 Vec deviation,
			 Vec delta,
			 GLDisplayList *dl)
{
  Num dev=deviation(0), del=delta(0);

  if (dev==0)
    {
      if (!exact_tesselations.insert(winding_rule, deviation(1), delta(1), dl))
	{
	  std::cerr << "WARNING: Exact interior display list not used.\n";
	  delete dl;
	}
      return;
    }

  Range range=insertEntry(dev, del);
  typedef Tesselations::iterator iterator;
  bool used=false;
  for (iterator it=range.first; it!=range.second; it++)
    if (it->insert(winding_rule, deviation(1), delta(1), dl)) used=true;
  if (used)
    // If the display list was used then 
    display_lists.push_back(dl);
  else
    {
      std::cerr << "WARNING: Interior display list not used.\n";
      delete dl;
    }
}

void
TesselationCache::insert(const StrokeStyle &style,
			 Vec deviation,
			 Vec delta,
			 GLDisplayList *dl)
{
  Num dev=deviation(0), del=delta(0);

  if (dev==0)
    {
      if (!exact_tesselations.insert(style, deviation(1), delta(1), dl))
	{
	  std::cerr << "WARNING: Exact interior display list not used.\n";
	  delete dl;
	}
      return;
    }

  Range range=insertEntry(dev, del);
  typedef Tesselations::iterator iterator;
  bool used=false;
  for (iterator it=range.first; it!=range.second; it++)
    if (it->insert(style, deviation(1), delta(1), dl)) used=true;
  if (used)
    // If the display list was used then 
    display_lists.push_back(dl);
  else
    {
      std::cerr << "WARNING: Interior display list not used.\n";
      delete dl;
    }
}

void
TesselationCache::clear()
{
  // Clear all caches
  while (!tesselations.empty())
    {
      tesselations.front().clear();
      tesselations.erase(tesselations.begin());
    }

  // Then remove all display lists referenced by the caches
  display_lists.clear();
}



// TesselationCache::TesselationCache2

TesselationCache::TesselationCache2::TesselationCache2()
  : scale(2),
    exact_tesselations(new CachedTesselation),
    central_deviation(0)
{
}

const CachedTesselation *
TesselationCache::TesselationCache2::renderEntry(Num d) const
{
  if (tesselations.empty())
    return NULL;

  // Find the largest deque index containing deviations which are all below
  // this delta
  int entry=0;
  if (central_deviation != 0 && d != 0)
    entry = (int)(floor(log(d/central_deviation)/log(scale)));

  // If the entry is out of bounds then return
  if (entry<0 || entry>=(int)tesselations.size())
    return NULL;

  return &tesselations[entry];
}

bool
TesselationCache::TesselationCache2::render(GLenum winding_rule,
					    Num &d) const
{
  // Try to find an exact tesselation if any are present
  if (exact_tesselations->render(winding_rule))
    return true;

  // If an exact tesselation could not be found then try to find an approximate
  // one
  const CachedTesselation *t=renderEntry(d);
  if (!t) return false;
  return t->render(winding_rule);
}

bool
TesselationCache::TesselationCache2::render(const StrokeStyle &style,
					    Num &d) const
{
  // Try to find an exact tesselation if any are present
  if (exact_tesselations->render(style))
    return true;

  // If an exact tesselation could not be found then try to find an approximate
  // one
  const CachedTesselation *t=renderEntry(d);
  if (!t) return false;
  return t->render(style);
}

int
TesselationCache::TesselationCache2::getIndex(Num d) const
{
  return int(floor(log(d/central_deviation)/log(scale)));
}

TesselationCache::TesselationCache2::Range
TesselationCache::TesselationCache2::insertEntry(Num dev,
						 Num del)
{
  if (tesselations.empty()) central_deviation = dev;

  int entry1=getIndex(dev), entry2=getIndex(del);
  while (entry1<0 || entry2<0)
    {
      entry1++;
      entry2++;
      tesselations.push_front(CachedTesselation());
      central_deviation /= scale;
    }
  while ((unsigned int)entry1>=tesselations.size() ||
	 (unsigned int)entry2>=tesselations.size())
    tesselations.push_back(CachedTesselation());
  return Range(tesselations.begin() + entry1,
	       ++tesselations.begin() + entry2);
}

bool
TesselationCache::TesselationCache2::insert(GLenum winding_rule,
					    Num dev,
					    Num del,
					    GLDisplayList *dl)
{
  if (dev==0)
    return exact_tesselations->insert(winding_rule, dl);

  Range range=insertEntry(dev, del);
  typedef Tesselations::iterator iterator;
  bool used=false;
  for (iterator it=range.first; it!=range.second; it++)
    if (it->insert(winding_rule, dl)) used=true;
  return used;
}

bool
TesselationCache::TesselationCache2::insert(const StrokeStyle &style,
					    Num dev,
					    Num del,
					    GLDisplayList *dl)
{
  if (dev==0)
    return exact_tesselations->insert(style, dl);

  Range range=insertEntry(dev, del);
  typedef Tesselations::iterator iterator;
  bool used=false;
  for (iterator it=range.first; it!=range.second; it++)
    if (it->insert(style, dl)) used=true;
  return used;
}

void
TesselationCache::TesselationCache2::clear()
{
  while (!tesselations.empty())
    {
      tesselations.front().clear();
      tesselations.erase(tesselations.begin());
    }
}



// Contour

Contour::Contour()
  : closed(true),
    start(Vec(0,0))
{
}

Contour::Contour(std::list<Segment *> &c)
  : std::list<Segment *>(c),
  closed(true),
  start(Vec(0,0))
{
}

Contour::Contour(const Vec s)
  : closed(false),
    start(s)
{
}

Contour::Contour(const Vec s,
		 std::list<Segment *> &c)
  : std::list<Segment *>(c),
  closed(false),
  start(s)
{
}

void
Contour::flatten(const Vec delta,
		 std::list<Vec> &contour,
		 Vec &max_deviation) const
{
  // OPTIM: Subsequent interior and stroke rendering will request similar
  // flattened contours (with the exception of open contours which aren't
  // required for interior rendering) so a cache may increase performance.
  if (empty())
    return;

  // Keep track of the starting vertex of the current segment
  Vec s;

  // If this contour is closed then the first vertex is the endpoint of the
  // last segment
  if (isClosed())
    s = back()->getEndPoint();
  else
    {
      s=getStart();
      contour.push_back(s);
    }

  // Loop through each segment in the input list and adaptively flatten them
  // to straight lines
  for (const_iterator it=begin(); it!=end(); it++)
    {
      Segment &seg=**it;
      seg.flatten(s, delta, contour, max_deviation);
      s = seg.getEndPoint();
    }
}

void
Contour::flatten(const ViewPort &viewport,
		 std::list<Vec> &contour,
		 Vec &max_deviation) const
{
  if (empty())
    return;

  // Keep track of the starting vertex of the current segment
  Vec s;

  // If this contour is closed then the first vertex is the endpoint of the
  // last segment
  if (isClosed())
    s = back()->getEndPoint();
  else
    {
      s=getStart();
      contour.push_back(s);
    }

  // Loop through each segment in the input list and adaptively flatten them
  // to straight lines
  for (const_iterator it=begin(); it!=end(); it++)
    {
      Segment &seg=**it;
      seg.flatten(s,
		  viewport,
		  contour,
		  max_deviation);
      s = seg.getEndPoint();
    }
}

const bool &
Contour::isClosed() const
{
  return closed;
}

void
Contour::setClosed(const bool c)
{
  // If we are opening an object then set the start point to the endpoint if
  // there is one
  if (closed && !c)
    if (!empty())
      start = back()->getEndPoint();
    else
      start = Vec(0, 0);
  closed = c;
}

const Vec &
Contour::getStart() const
{
  if (!isClosed())
    return start;

  std::cerr << "ERROR: Start point of a closed path was requested.\n";
  throw;
  return start;
}

void
Contour::setStart(const Vec &s)
{
  if (!isClosed())
    start = s;

  std::cerr << "ERROR: Tried to set the start point of a closed path.\n";
  throw;
  return;
}

const Bound &
Contour::getBound() const
{
  return bound;
}

const Bound &
Contour::calcBound() const
{
  if (empty())
    return bound=Bound();

  Vec s=isClosed() ? back()->getEndPoint() : getStart();

  const_iterator it=begin();
  {
    Segment &seg=**it;
    bound = seg.calcBound(s);
    s = seg.getEndPoint();
    it++;
  }
  for (; it!=end(); it++)
    {
      Segment &seg=**it;

      bound = Bound(bound, seg.calcBound(s));

      s = seg.getEndPoint();
    }

  return bound;
}

void
Contour::save(std::ostream &out,
	      const Store *store) const
{
  out << "contour ";
  for (const_iterator it=begin(); it!=end(); it++)
    out //<< store.segmentlist.getIndex(*it)
	<< "\n";
}



// ContourGeometry

ContourGeometry::ContourGeometry()
{
}

ContourGeometry::~ContourGeometry()
{
  tesselation_cache.clear();
}

void
ContourGeometry::flatten(Vec delta,
			 std::list<std::list<Vec> > &polygon,
			 Vec &max_deviation,
			 bool only_closed) const
{
  for (const_iterator it=begin(); it!=end(); it++)
    {
      const Contour &contour=*it;
      
      if ((only_closed && contour.isClosed()) || !only_closed)
	{
	  polygon.push_back(std::list<Vec>());
	  contour.flatten(delta,
			  polygon.back(),
			  max_deviation);
	}
    }
}

void
ContourGeometry::flatten(ViewPort viewport,
			 std::list<std::list<Vec> > &polygon,
			 Vec &max_deviation,
			 bool only_closed) const
{
  for (const_iterator it=begin(); it!=end(); it++)
    {
      const Contour &contour=*it;
      
      if ((only_closed && contour.isClosed()) || !only_closed)
	{
	  polygon.push_back(std::list<Vec>());
	  contour.flatten(viewport,
			  polygon.back(),
			  max_deviation);
	}
    }
}

void
ContourGeometry::renderSegmentStyle(const ViewPort &viewport,
				    const Vec &r1,
				    const Vec &r2,
				    const bool isclosed,
				    const StrokeStyle &style) const
{
  // Draw the line segment
  style.renderSegment(viewport, r1, r2);

  if (isclosed)
    {
      // Draw the joins (rounded joins will look like rounded endcaps
      style.renderJoin(viewport, r1, r2, r1);
      style.renderJoin(viewport, r2, r1, r2);
    }
  else
    {
      // Draw the end caps
      style.renderEndCap(viewport, 1, r1, r2);
      style.renderEndCap(viewport, 2, r2, r1);
    }
}

void
ContourGeometry::renderStrokeStyle(const ViewPort &viewport,
				   const std::list<Vec> &contour,
				   const bool isclosed,
				   const StrokeStyle &style) const
{
  std::list<Vec>::const_iterator
    it1=contour.begin(),
    it2,
    it3;

  // If the contour is empty then exit
  if (it1 == contour.end())
    return;

  it2=it1;
  it2++;

  // If the contour has only one vertex then exit
  if (it2 == contour.end())
    return;

  it3=it2;
  it3++;

  // If the contour has two vertices then draw a single line segment with its
  // end or join caps
  if (it3 == contour.end())
    renderSegmentStyle(viewport, *it1, *it2, isclosed, style);

  // Draw the first line segment
  style.renderSegment(viewport, *it1, *it2);

  // Draw the first join or end cap
  if (isclosed)
    // First join
    style.renderJoin(viewport, contour.back(), *it1, *it2);
  else
    // First end cap
    style.renderEndCap(viewport, 1, *it1, *it2);

  // Draw join and line segments up to the end
  while (it3 != contour.end())
    {
      style.renderJoin(viewport, *it1, *it2, *it3);
      style.renderSegment(viewport, *it2, *it3);
      it1 = it2;
      it2 = it3;
      it3++;
    }

  if (isclosed)
    {
      // Line segment and join back to the start
      it3 = contour.begin();
      style.renderJoin(viewport, *it1, *it2, *it3);
      style.renderSegment(viewport, *it2, *it3);
    }
  else
    // Last end cap
    style.renderEndCap(viewport, 2, *it3, *it2);
}

Vec
ContourGeometry::renderNewStrokes(StrokeStyle style,
				  Vec delta) const
{
  Vec deviation(0, 0);

  std::list<std::list<Vec> > polygon;

  flatten(delta,
	  polygon,
	  deviation,
	  false);

  std::list<std::list<Vec> >::iterator pit=polygon.begin();

  // FIXME: Stroking should also alter the deviation
  for (const_iterator it=begin(); it!=end(); it++, pit++)
    {
      std::list<Vec> &contour=*pit;
      // FIXME: We do not wish to remove coincident vertices as this changes
      // the result (e.g. for stroking). We should instead make the stroking
      // code numerically stable so that this code can be removed
      /*
      if (contour.size()>1)
	for (std::list<Vec>::iterator it1=contour.begin(), it2=it1++;
	     it2!=contour.end();)
	  if (*it1==*it2)
	    {
	      std::list<Vec>::iterator t=it2;
	      it2++;
	      contour.erase(t);
	      std::cerr << "WARNING: Coincident vertex.\n";
	    }
	  else
	    {
	      it1++;
	      it2++;
	    }
      */
      renderStrokeStyle(ViewPort(getBound(), delta),
			contour,
			it->isClosed(),
			style);
    }

  return deviation;
}

void
ContourGeometry::renderStrokes(StrokeStyle style,
			       Vec delta) const
{
  if (!tesselation_cache.render(style,
				delta))
    {
      Vec deviation(0, 0);
      GLDisplayList &dl=*(new GLDisplayList());
      dl.begin();
      deviation = renderNewStrokes(style, delta);
      dl.end();
      tesselation_cache.insert(style, deviation, delta, &dl);
    }
}

Vec
ContourGeometry::renderStrokes(StrokeStyle style,
			       ViewPort viewport) const
{
  Vec deviation(0, 0);

  std::list<std::list<Vec> > polygon;

  flatten(viewport,
	  polygon,
	  deviation,
	  false);

  std::list<std::list<Vec> >::const_iterator pit=polygon.begin();

  for (const_iterator it=begin(); it!=end(); it++, pit++)
    renderStrokeStyle(viewport,
		      *pit,
		      it->isClosed(),
		      style);

  return deviation;
}

Vec
ContourGeometry::renderNewInterior(GLenum winding_rule,
				   Vec delta) const
{
  Vec deviation(0, 0);
  std::list<std::list<Vec> > polygon;

  // OPTIM: Try culling sequential segments which are outside the viewport in
  // the same direction.
  flatten(delta,
	  polygon,
	  deviation,
	  true);

  bool stencil=false;
  if (stencil)
    {
      // Render using the stencil buffer to keep track of winding numbers on a
      // per pixel basis. Only a 1-bit stencil buffer is needed because the
      // non-zero winding rule can just set the stencil buffer to 1 and the
      // even-odd winding rule can repeatedly invert the current value.
      // Testing this code, the stencil buffer approach to winding appears to
      // be considerably slower.
      glEnable(GL_STENCIL_TEST);
      glDrawBuffer(GL_NONE);

      glStencilFunc(GL_ALWAYS, 0, 0);
      glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);
      
      glBegin(GL_QUADS);
      const Bound &b=getBound();
      const Vec &l=b.l, &u=b.u;
      glVertex(l);
      glVertex(Vec(l(0), u(1)));
      glVertex(u);
      glVertex(Vec(u(0), l(1)));
      glEnd();

      glEnable(GL_STENCIL_TEST);
      glStencilFunc(GL_ALWAYS, 1, 1);
      if (winding_rule == GLU_TESS_WINDING_ODD)
	glStencilOp(GL_KEEP, GL_KEEP, GL_INVERT);
      else
	glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);

      for (std::list<std::list<Vec> >::iterator it1=polygon.begin(); it1!=polygon.end();
	   it1++)
	{
	  std::list<Vec> &l=*it1;
	  glBegin(GL_TRIANGLE_FAN);
	  for (std::list<Vec>::iterator it2=l.begin(); it2!=l.end(); it2++)
	    glVertex(*it2);
	  glEnd();
	}
      
      glStencilFunc(GL_NOTEQUAL, 0, 1);
      glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
      glDrawBuffer(GL_BACK);
      
      glBegin(GL_QUADS);
      glVertex(l);
      glVertex(Vec(l(0), u(1)));
      glVertex(u);
      glVertex(Vec(u(0), l(1)));
      glEnd();
      
      glDisable(GL_STENCIL_TEST);
    }
  else
    // Render using the GLU polygon tesselator to generate a set of non-
    // overlapping triangles which represent the interior of this polygon
    renderTesselation(polygon, winding_rule);

  return deviation;
}

void
ContourGeometry::renderInterior(GLenum winding_rule,
				Vec delta) const
{
  if (!tesselation_cache.render(winding_rule,
				delta))
    {
      Vec deviation(0, 0);
      GLDisplayList &dl=*(new GLDisplayList());
      dl.begin();
      deviation = renderNewInterior(winding_rule, delta);
      dl.end();
      tesselation_cache.insert(winding_rule, deviation, delta, &dl);
    }
}

Vec
ContourGeometry::renderInterior(GLenum winding_rule,
				ViewPort viewport) const
{
  Vec deviation(0, 0);

  std::list<std::list<Vec> > polygon;

  flatten(viewport,
	  polygon,
	  deviation,
	  true);

  renderTesselation(polygon, winding_rule);

  return deviation;
}

const Bound &
ContourGeometry::getBound() const
{
  return bound;
}

const Bound &
ContourGeometry::calcBound() const
{
  if (empty())
    return bound = Bound();
  const_iterator it=begin();
  {
    bound = it->calcBound();
    it++;
  }
  for (; it!=end(); it++)
    bound = Bound(bound, it->calcBound());
  return bound;
}

void
ContourGeometry::clearCache() const
{
}
