/** file: geometry.h
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

class CachedTesselation;
class CachedPartialTesselation;
class Contour;
class ContourGeometry;

#ifndef GEOMETRY_H
#define GEOMETRY_H

#include "caches.h"
#include "tree.h"
#include "bound.h"
#include "autoptr.h"
#include "gldisplaylist.h"
#include "vecmat.h"
#include "stroke.h"
#include "segments.h"

#include <GL/glu.h>
#include <memory>
#include <deque>
#include <list>
#include <map>



/** Cached tesselations of geometries at different deviations.
 *
 *  Exact and approximate tesselations of interiors and strokes of geometries
 *  can be cached by this class.
 *
 *  Internally, exact tesselations are stored separately from approximate
 *  tesselations. Approximate tesselations are stored in a deque of deques of
 *  CachedTesselation classes. The first deque orders by x deviation, the
 *  second by y deviation.
 */
class TesselationCache
{
 private:
  class TesselationCache2
    {
    private:
      Num scale;

      /** Separate cache for exact tesselations. */
      AutoPtr<CachedTesselation> exact_tesselations;

      /** Type of container used to store tesselations in this cache. */
      typedef std::deque<CachedTesselation> Tesselations;

      /** Cached tesselation container. */
      Tesselations tesselations;

      /** Range of iterators into the tesselation cache. */
      typedef std::pair<Tesselations::iterator, Tesselations::iterator> Range;

      /** Maximum deviation represented by the first entry in the container. */
      Num central_deviation;

      int
	getIndex(Num d) const;

      Range
	insertEntry(Num dev,
		    Num del);

      const CachedTesselation *
	renderEntry(Num d) const;

    public:
      TesselationCache2();

      bool
	insert(GLenum winding_rule,
	       Num dev,
	       Num del,
	       GLDisplayList *dl);
      
      bool
	insert(const StrokeStyle &style,
	       Num dev,
	       Num del,
	       GLDisplayList *dl);

      bool
	render(GLenum winding_rule,
	       Num &d) const;

      bool
	render(const StrokeStyle &style,
	       Num &d) const;

      void
	clear();
    };

  Num scale;

  /** List of display lists which we are responsible for. */
  std::list<AutoPtr<GLDisplayList> > display_lists;

  /** Separate cache for exact tesselations. */
  TesselationCache2 exact_tesselations;

  /** Type of container used to store tesselations in this cache. */
  typedef std::deque<TesselationCache2> Tesselations;

  /** Cached tesselation container. */
  Tesselations tesselations;

  /** Range of iterators into the tesselation cache. */
  typedef std::pair<Tesselations::iterator, Tesselations::iterator> Range;

  /** Maximum deviation represented by the first entry in the container. */
  Num central_deviation;

  int
    getIndex(Num d) const;
  
  Range
    insertEntry(Num dev,
		Num del);

  const TesselationCache2 *
    renderEntry(Num d) const;

 public:
  TesselationCache();

  bool
    render(GLenum winding_rule,
	   Vec delta) const;

  bool
    render(const StrokeStyle &style,
	   Vec delta) const;

  void
    insert(GLenum winding_rule,
	   Vec deviation,
	   Vec delta,
	   GLDisplayList *dl);

  void
    insert(const StrokeStyle &style,
	   Vec deviation,
	   Vec delta,
	   GLDisplayList *dl);

  void
    clear();
};



/** An open or closed path.
 *
 *  A linked list of path segments which are optionally closed. In the case of
 *  an open path, the first vertex is stored in the member vector `start'.
 */
class Contour : public std::list<Segment *>
{
 private:
  /** Only a closed vector path can be filled. */
  bool closed;

  /** Start point for this contour if it is open. */
  Vec start;

  mutable Bound bound;

 public:
  /** Contruct a closed, empty contour. */
  Contour();

  /** Contruct a closed contour. */
  Contour(std::list<Segment *> &c);

  /** Contruct an open contour with the given start vertex. */
  Contour(const Vec s);

  /** Contruct an open contour with the given start vertex. */
  Contour(const Vec s,
	  std::list<Segment *> &c);

  void
    flatten(const Vec delta,
	    std::list<Vec> &contour,
	    Vec &max_deviation) const;
  
  void
    flatten(const ViewPort &viewport,
	    std::list<Vec> &contour,
	    Vec &max_deviation) const;

  const bool &
    isClosed() const;

  /** Close or open this contour.
   *
   *  If the contour was previously closed and this call sets the contour to
   *  be open then the starting vertex is initialised to the last vertex. */
  void
    setClosed(const bool c);

  /** Get the starting vertex of this path. */
  const Vec &
    getStart() const;

  void
    setStart(const Vec &s);

  const Bound &
    getBound() const;

  const Bound &
    calcBound() const;

  void
    save(std::ostream &out,
	 const Store *store) const;
};



/** Geometry of a contour object.
 *
 *  Multiple contour objects are allowed to reference a single contour
 *  geometry. This is useful when the same geometry appears many times in the
 *  scene with different fill types, winding rules or line styles. This is also
 *  useful for font rendering as the scene's font cache can generate multiple
 *  references to the geometry of each glyph.
 *
 *  The geometry of the interior of a contour object is defined by the contours
 *  and winding rule. The geometry of the strokes of a contour object are
 *  defined by the contours and the stroke style.
 *
 *  This class represents the interior and stroke geometries of its contours.
 *  Rendering performance is improved by caching tesselations of the geometry
 *  at different deviations.
 */
class ContourGeometry : public std::list<Contour>
{
 private:
  /** Bounding area of this geometry's contours. */
  mutable Bound bound;

  /** Cached tesselations of interiors and strokes. */
  mutable TesselationCache tesselation_cache;

  /** Flatten the contours into the given polygon. */
  void
    flatten(const Vec delta,
	    std::list<std::list<Vec> > &polygon,
	    Vec &max_deviation,
	    bool only_closed) const;

  /** Flatten the contours into the given polygon, culling path segments
   *  outside the given viewport. */
  void
    flatten(ViewPort viewport,
	    std::list<std::list<Vec> > &polygon,
	    Vec &max_deviation,
	    bool only_closed) const;

  /** Render the stroke of a flattened contour which has only two segments. */
  void
    renderSegmentStyle(const ViewPort &viewport,
		       const Vec &r1,
		       const Vec &r2,
		       const bool isclosed,
		       const StrokeStyle &style) const;

  /** Render the stroke of a flattened contour. */
  void
    renderStrokeStyle(const ViewPort &viewport,
		      const std::list<Vec> &contour,
		      const bool isclosed,
		      const StrokeStyle &style) const;

 public:
  /** Contruct and empty contour geometry. */
  ContourGeometry();

  ~ContourGeometry();

  /** Flatten the contours and stroking and rendering the result. */
  Vec
    renderNewStrokes(StrokeStyle style,
		     Vec delta) const;

  /** Render the strokes of the contours.
   *
   *  First try to render an existing cache entry which has a deviation smaller
   *  than the requested delta. If one could not be found then flatten the
   *  contours using the delta and insert the rendered result into the cache.
   */
  void
    renderStrokes(StrokeStyle style,
		  Vec delta) const;

  /** Flatten the contours, culling those outside the viewport, and stroke and
   *  render the result. */
  Vec
    renderStrokes(StrokeStyle style,
		  ViewPort viewport) const;

  /** Flatten the contours and tesselating and rendering the result. */
  Vec
    renderNewInterior(GLenum winding_rule,
		      Vec delta) const;

  /** Flatten the contours and render their interior with the given winding
   *  rule. */
  Vec
    renderInterior(GLenum winding_rule,
		   ViewPort viewport) const;

  /** Flatten the contours and render their interior using the entire
   *  geometry. */
  void
    renderInterior(GLenum winding_rule,
		   Vec delta) const;

  const Bound &
    getBound() const;

  const Bound &
    calcBound() const;

  /** Remove all cached geometries. */
  void
    clearCache() const;
};



#endif /* GEOMETRY_H */
