/* file: load.cc
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "scene.h"
#include "fill_types.h"
#include "num.h"
#include "vecmat.h"
#include "fill_types.h"

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <string>



void
discard_char(std::istream &in)
{
  char c;
  in.get(c);
}

void
loadObject(std::istream &in,
	   Store &store,
	   ContourObject &o)
{
  {
    std::string s;
    in >> s;
    if (s != std::string("filltypes"))
      {
	std::cerr << "ERROR: Expected `filltypes' in contour object, got `"
		  << s << "'.\n";
	throw;
      }
    while (in.peek()==' ' || in.peek()==10)
      discard_char(in);
  }

  // Load fill types
  while (!in.eof() && in.peek()>='0' && in.peek()<='9')
    {
      int i;
      in >> i;
      o.fill_types.push_back(store.fill_types[i]);
    }
  if (!o.fill_types.empty())
    discard_char(in);

  {
    std::string s;
    in >> s;
    if (s != std::string("linestyles"))
      {
	std::cerr << "ERROR: Expected `linestyles' in contour object, got `"
		  << s << "'.\n";
	throw;
      }
    while (in.peek()==' ' || in.peek()==10)
      discard_char(in);
  }

  // Load line styles
  while (!in.eof() && in.peek()>='0' && in.peek()<='9')
    {
      int i;
      in >> i;
      o.line_styles.push_back(store.line_styles[i]);
    }
  if (!o.line_styles.empty())
    discard_char(in);

  ContourGeometry &g=o.getGeometry();

  while (!in.eof() && in.peek() != 10)
    {
      std::string t;
      in >> t;
      if (t==std::string("Contour"))
	{
	  discard_char(in);
	  if (in.peek() != 10)
	    {
	      std::string o;
	      in >> o;
	      Vec s;
	      in >> s;
	      s(1)=-s(1);
	      g.push_back(Contour(s));
	      discard_char(in);
	    }
	  else
	    {
	      g.push_back(Contour());
	      discard_char(in);
	    }
	  while (!in.eof() && in.peek() != 10)
	    {
	      std::string t;
	      in >> t;
	      if (t==std::string("Line"))
		{
		  Vec e;
		  in >> e;
		  e(1)=-e(1);
		  store.segments.push_back(new LineSegment(e));
		}
	      else
		if (t==std::string("Bezier"))
		  {
		    Vec c1, c2, e;
		    in >> c1 >> c2 >> e;
		    c1(1)=-c1(1);
		    c2(1)=-c2(1);
		    e(1)=-e(1);
		    store.segments.push_back(new BezierSegment(c1, c2, e));
		  }
		else
		  {
		    std::cerr << "ERROR: Unknown segment type `"
			      << t << "'.\n";
		    throw;
		  }
	      g.back().push_back(store.segments.back());
	      discard_char(in);
	    }
	}
      else
	{
	  std::cerr << "ERROR: Unknown geometry element `" << t << "'.\n";
	  throw;
	}
    }
  discard_char(in);
}

void
load_scene(const char* filename,
	   Scene &scene)
{
  // Create a new, empty group of objects
  Group &root=scene.getRoot();

  Store &store=scene.getStore();

  std::ifstream in(filename);

  {
    std::string s;
    in >> s;
    if (s != std::string("FillTypes"))
      {
	std::cerr << "ERROR: Expected `FillTypes', got `" << s << "'.\n";
	throw;
      }
    discard_char(in);
  }

  // Load fill types
  while (!in.eof() && in.peek() != 10)
    {
      std::string t;
      in >> t;
      discard_char(in);
      if (t==std::string("FlatFill"))
	  {
	    Colour c;
	    in >> c;
	    store.fill_types.push_back(new FlatFill(c));
	    discard_char(in);
	  }
      if (t==std::string("GradientFill"))
	  {
	    // FIXME: Gradient fill loading not yet implemented
	  }
    }
  discard_char(in);

  {
    std::string s;
    in >> s;
    if (s != std::string("LineStyles"))
      {
	std::cerr << "ERROR: Expected `LineStyles', got `" << s << "'.\n";
	throw;
      }
    discard_char(in);
  }

  // Load line styles
  while (!in.eof() && in.peek() != 10)
    {
      int fill;
      in >> fill;
      double width;
      in >> width;

      // FIXME: Join and endcap style loading not yet implemented
      std::string join;
      in >> join;
      std::string endcaps;
      in >> endcaps;

      store.line_styles.push_back(new LineStyle(StrokeStyle(width,
							    store.roundjoin,
							    store.noendcap,
							    store.noendcap)));
      store.line_styles.back()->fill_types.push_back(store.fill_types[fill]);
      discard_char(in);
    }
  discard_char(in);

  {
    std::string s;
    in >> s;
    if (s != std::string("Scene"))
      {
	std::cerr << "ERROR: Expected `Scene', got `" << s << "'.\n";
	throw;
      }
    discard_char(in);
    discard_char(in);
  }

  // Load objects
  while (!in.eof() && in.peek() != 10)
    {
      std::string t;
      in >> t;
      if (t==std::string("ContourObject"))
	{
	  ContourObject &o=*new ContourObject(&store, GLU_TESS_WINDING_ODD);
	  root.push_front(&o);
	  loadObject(in, store, o);
	}
      discard_char(in);
    }
  discard_char(in);

  scene.calcBounds();
}
