/** file: main.cc
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/



#include "scene.h"
#include "viewport.h"
#include "fill_types.h"
#include "num.h"
#include "vecmat.h"
#include "usec_timer.h"
#include "test.h"
#include "load.h"

#include <cmath>
#include <cstdlib>
#include <iostream>
#include <string>
#include <GL/gl.h>
#include <GL/glu.h>
#include <SDL/SDL.h>

using namespace std;		// Lots of IO

ViewPort *viewport;
bool
fsantialiasing=false,
  edge_antialiasing=false,
  wireframe=false,
  randomised_viewport=false,
  demo=false;
Num size=.5;
Vec centre(.5, .5);
int load_external_scene = 0;
long long starttime=gettime(), oldtime=starttime;
char external_scene_file[80]; // buffer overflow

void
resizeScene(const int Width,
	    const int Height);

void
toggleFullScreen();

void
toggleGrab();

bool
handleEvent(SDL_Event *event);

int
runSmoke(int argc,
	 char *argv[],
	 float gamma);

int Window_Width=1200;
int Window_Height=1200;
long long frame_last_clock=0;
int frame_counter=0;
bool depthflag=false;

void
resizeScene(const int Width,
	    const int Height)
{
  // Set the viewport to cover the new window size.
  glViewport(0, 0, Width, Height);

  // Keep a local copy of the extent of the screen.
  Window_Width  = Width;
  Window_Height = Height;

  glClearDepth(1.0);
  glClear(GL_COLOR_BUFFER_BIT);
  depthflag = false;

  if (demo)
    {
      long long currenttime=gettime()-starttime;
      if (currenttime<5e6)
	{
	  double p=currenttime/5.0e6;

	  centre = (1-p)*Vec(67.7, -94.9) + p*Vec(45.8, -23.0);
	  //size = (1-p)*1000 + p*15;
	  size = 1000*pow(15./1000., p);
	}
      else
	if (currenttime<10e6)
	  {
	    double p=(currenttime - 5.0e6)/5.0e6;

	    centre = (1-p)*Vec(45.8, -23.0) + p*Vec(-166.9, -101.692);
	    //size = (1-p)*15 + p*50;
	    size = 15*pow(50./15., p);
	  }
	else
	  if (currenttime<15e6)
	    {
	      double p=(currenttime - 10.0e6)/5.0e6;

	      centre = (1-p)*Vec(-166.9, -101.692) + p*Vec(67.7, -94.9);
	      //size = (1-p)*50 + p*300;
	      size = 50*pow(300./50., p);
	    }
	  else
	    if (currenttime<20e6)
	      {
		double p=(currenttime - 15.0e6)/5.0e6;
		
		centre = (1-p)*Vec(67.7, -94.9) + p*Vec(45.8, -23.0);
		//size = (1-p)*1000 + p*15;
		size = 300*pow(15./300., p);
	      }
    }

  // Create a new viewport to render the objects with
  delete viewport;
  viewport = new ViewPort(Bound(centre-Vec(size, size),
				centre+Vec(size, size)),
			  Window_Width,
			  Window_Height);

  // Render the scene
  try
    {
      if (fsantialiasing)
	scene->renderAA(Window_Width,
			Window_Height,
			*viewport);
      else
	scene->render(*viewport);
    }
  catch (...)
    {
      cerr << "Exception caught.\n";
    }

  if (!fsantialiasing)
    SDL_GL_SwapBuffers();
}

void
toggleFullScreen()
{
  SDL_Surface *screen;
  
  screen = SDL_GetVideoSurface();

  if ( SDL_WM_ToggleFullScreen(screen) )
    cout << "Toggled fullscreen mode - now " << ((screen->flags&SDL_FULLSCREEN) ? "fullscreen" : "windowed") << "\n";
  else
    cout << "Unable to toggle fullscreen mode\n";

  resizeScene(Window_Width, Window_Height);
}

void
toggleGrab()
{
  SDL_GrabMode mode;
  
  cout << "Ctrl-G: toggling input grab!\n";

  mode = SDL_WM_GrabInput(SDL_GRAB_QUERY);

  if (mode == SDL_GRAB_ON)
    cout << "Grab was on\n";
  else
    cout << "Grab was off\n";

  mode = SDL_WM_GrabInput(mode ? SDL_GRAB_OFF : SDL_GRAB_ON);

  if (mode == SDL_GRAB_ON)
    cout << "Grab is now on\n";
  else
    cout << "Grab is now off\n";
}

enum AppState { STOPPED, ZOOMING_IN, ZOOMING_OUT };

AppState appState = STOPPED;

int totalFrames = 0; // just for the zoom
long long zoomStart = 0;
long long zoomEnd = 0;

void 
doZoomFrameStats()
{
  cout << "Displayed " << totalFrames << " in " << (zoomEnd - zoomStart);
  cout << " us == " << ( (float)(zoomEnd - zoomStart) / 1000.0 ) << "ms => ";
  cout << 1000000.0 * (float)totalFrames / (float)(zoomEnd - zoomStart)
       << " fps.\n";
}

void
updateState()
{
  long long newtime=gettime();

  switch (appState)
    {
    case STOPPED:
      break;
    case ZOOMING_IN:
      {
	size /= pow(1.01, (newtime-oldtime)*4e-5);
	totalFrames++;
	resizeScene(Window_Width, Window_Height);
	break;
      }
    case ZOOMING_OUT:
      {
	size *= pow(1.01, (newtime-oldtime)*4e-5);
	totalFrames++;
	resizeScene(Window_Width, Window_Height);
	break;
      }
    };

  oldtime = newtime;
}

bool
handleEvent(SDL_Event *event)
{
  bool done;

  done = false;

  switch( event->type )
    {
    case SDL_ACTIVEEVENT:
      {
	resizeScene(Window_Width, Window_Height);
	/* See what happened */
	/*
	cout << "app " << (event->active.gain ? "gained" : "lost") << endl;
	if (event->active.state & SDL_APPACTIVE)
	  cout << "active ";
	else
	  if (event->active.state & SDL_APPMOUSEFOCUS)
	    cout << "mouse ";
	  else
	    if (event->active.state & SDL_APPINPUTFOCUS)
	      cout << "input ";
	cout << "focus\n";
	*/
	break;
      }

    case SDL_KEYDOWN:
      {
	if ( event->key.keysym.sym == SDLK_ESCAPE )
	  done = true;
	if ( (event->key.keysym.sym == SDLK_g) &&
	     (event->key.keysym.mod & KMOD_CTRL) )
	  toggleGrab();
	if ( (event->key.keysym.sym == SDLK_RETURN) &&
	     (event->key.keysym.mod & KMOD_ALT) )
	  toggleFullScreen();
	cout << "key '" << SDL_GetKeyName(event->key.keysym.sym) << "' pressed\n";
	break;
      }

    case SDL_MOUSEBUTTONDOWN:
      {
	const int
	  w=Window_Width,
	  h=Window_Height;
	Vec p(Num(event->button.x)/w,
	      1 - Num(event->button.y)/h);
	const Bound &b=viewport->getBound();
	centre = p*(b.u-b.l)+b.l;
	switch (event->button.button)
	  {
	  case 1:
	    {
	      cerr << "Starting continuous zoom in to " << centre << "\n";
	      appState = ZOOMING_IN;
	      totalFrames = 0;
	      zoomStart = gettime();
	      break;
	    }
	  case 3:
	    {
	      cerr << "Starting continuous zoom out to " << centre << "\n";
	      appState = ZOOMING_OUT;
	      totalFrames = 0;
	      zoomStart = gettime();
	      break;
	    }
	  };
	resizeScene(Window_Width, Window_Height);
	break;
      }
    case SDL_MOUSEBUTTONUP:
      {
	appState = STOPPED;
	zoomEnd = gettime();
	doZoomFrameStats();
	break;
      }
    case SDL_QUIT:
      {
	done = true;
	break;
      }
    }

  return done;
}


int
runSmoke(int argc,
	 char *argv[],
	 float gamma)
{
  int i=0, bpp=8, frames=0;
  Uint32 video_flags=SDL_OPENGL;
  
  if(SDL_Init(SDL_INIT_VIDEO) < 0)
    {
      cerr << "ERROR: Couldn't initialize SDL: " << SDL_GetError() << "\n";
      exit(1);
    }
  else
    cout << "Initialised SDL.\n";

  for ( i=1; argv[i]; ++i )
    {
      if ( strcmp(argv[i], "-fullscreen") == 0 ) 
	{
	  video_flags |= SDL_FULLSCREEN;
	  continue;
	}
      if ( strcmp(argv[i], "-loadscene") == 0 )
	{
	  strcpy(external_scene_file, argv[i+1]);
	  load_external_scene = 1;
	  size=260;
	  centre = Vec(67.7,-110);
	}
      if ( strcmp(argv[i], "-demo") == 0 )
	demo=true;
    }

  cout << "Setting SDL attributes...\n";
  SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 1);
  SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 1);
  SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 1);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 1);
  SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 1);
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
  SDL_GL_SetAttribute(SDL_GL_ACCUM_RED_SIZE, 0);
  SDL_GL_SetAttribute(SDL_GL_ACCUM_GREEN_SIZE, 0);
  SDL_GL_SetAttribute(SDL_GL_ACCUM_BLUE_SIZE, 0);
  SDL_GL_SetAttribute(SDL_GL_ACCUM_RED_SIZE, 0);
  SDL_GL_SetAttribute(SDL_GL_ACCUM_GREEN_SIZE, 0);
  SDL_GL_SetAttribute(SDL_GL_ACCUM_BLUE_SIZE, 0);

  {
    bool aaerror=false;
    if (fsantialiasing && ((Window_Width-1) & Window_Width) != 0)
      {
	aaerror=true;
	Window_Width=256;
      }
    
    if (fsantialiasing && ((Window_Height-1) & Window_Height) != 0)
      {
	aaerror=true;
	Window_Height=256;
      }

    if (aaerror)
      cerr << "Antialiasing is currently only supported in integer power of two image sizes -> antialiasing disabled.\n";
  }

  if (SDL_SetVideoMode(Window_Width, Window_Height, bpp, video_flags) == NULL)
    {
      cerr << "Couldn't set GL mode: " << SDL_GetError() << endl;
      SDL_Quit();
      exit(1);
    }
  else
    cout << "Set GL mode.\n";

  cout << "Screen BPP : " << int(SDL_GetVideoSurface()->format->BitsPerPixel) << endl
            << "Vendor     : " << glGetString(GL_VENDOR) << endl
            << "Renderer   : " << glGetString(GL_RENDERER) << endl
            << "Version    : " << glGetString(GL_VERSION) << endl
            << "Extensions : " << glGetString(GL_EXTENSIONS) << endl
            << endl;

  int value=0;
  cout << "Reading SDL attributes:\n" << flush;
  SDL_GL_GetAttribute(SDL_GL_RED_SIZE, &value);
  cout << "SDL_GL_RED_SIZE: requested " << 1 << ", got " << value << "\n";
  SDL_GL_GetAttribute(SDL_GL_GREEN_SIZE, &value);
  cout << "SDL_GL_GREEN_SIZE: requested " << 1 << ", got " << value << "\n";
  SDL_GL_GetAttribute(SDL_GL_BLUE_SIZE, &value);
  cout << "SDL_GL_BLUE_SIZE: requested " << 1 << ", got " << value << "\n";
  SDL_GL_GetAttribute(SDL_GL_DEPTH_SIZE, &value);
  cout << "SDL_GL_DEPTH_SIZE: requested " << 1 << ", got " << value << "\n";
  SDL_GL_GetAttribute(SDL_GL_STENCIL_SIZE, &value);
  cout << "SDL_GL_STENCIL_SIZE: requested " << 1 << ", got " << value << "\n";
  SDL_GL_GetAttribute(SDL_GL_DOUBLEBUFFER, &value);
  cout << "SDL_GL_DOUBLEBUFFER: requested 1, got " << value << "\n";
  SDL_GL_GetAttribute(SDL_GL_ACCUM_RED_SIZE, &value);
  cout << "SDL_GL_ACCUM_RED_SIZE: requested " << 0 << ", got " << value << "\n";
  SDL_GL_GetAttribute(SDL_GL_ACCUM_GREEN_SIZE, &value);
  cout << "SDL_GL_ACCUM_GREEN_SIZE: requested " << 0 << ", got " << value << "\n";
  SDL_GL_GetAttribute(SDL_GL_ACCUM_BLUE_SIZE, &value);
  cout << "SDL_GL_ACCUM_BLUE_SIZE: requested " << 0 << ", got " << value << "\n";

  /* Set the window manager title bar */
  cout << "Setting SDL caption.\n" << flush;
  SDL_WM_SetCaption("Smoke Vector Graphics", "smoke");

  /* Set the gamma for the window */
  if (gamma != 0.0)
    {
      cout << "Setting gamma correction.\n" << flush;
      SDL_SetGamma(gamma, gamma, gamma);
    }

  frames = 0;
  if (load_external_scene == 1)
    {
      cout << "Attempting to load an external scene from "
	   << external_scene_file << "\n" << flush;
      scene = new Scene;
      load_scene(external_scene_file, *scene);
    }
  else
    {
      cout << "Using built-in scene.\n" << flush;
      initScene();
    }

  glClearColor(1, 1, 1, 0);
  //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  //glPolygonMode(GL_FRONT, GL_LINE);
  //glPolygonMode(GL_BACK, GL_LINE);
  //glEnable(GL_BLEND);
  //glEnable(GL_POINT_SMOOTH);
  //glEnable(GL_LINE_SMOOTH);
  //glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
  //glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
  if (edge_antialiasing)
    {
      glEnable(GL_POLYGON_SMOOTH);
      glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
    }

  resizeScene(Window_Width, Window_Height);

  for (int i=1; i<7; i++)
    scene->render(ViewPort(viewport->getBound(),
			   viewport->getDelta()/Num(1<<i)));

  if (demo)
    {
      bool done = false;
      int imax=10;
      
      cerr << "{";
      
      for (i=imax; i<=imax && !done; i++)
	{
	  int dim=int(sqrt(i*144000.)+.5);

	  //Window_Width=Window_Height=dim;

	  cerr << "{" << dim*dim << "," << flush;

	  if (SDL_SetVideoMode(Window_Width,
			       Window_Height,
			       bpp, video_flags) == NULL)
	    {
	      cerr << "Failed to set GL video mode.\n";
	      done=true;
	      continue;
	    }

	  frame_counter=0;
	  starttime=gettime();
	  
	  while(!done && gettime()-starttime < 20e6)
	    {
	      GLenum gl_error;
	      char* sdl_error;
	      SDL_Event event;
	      
	      /* Check for error conditions. */
	      gl_error = glGetError( );
	      
	      if(gl_error != GL_NO_ERROR)
		cerr << "ERROR: OpenGL error: '" << gl_error << "': "
		     << gluErrorString(gl_error) << "\n";
	      
	      sdl_error = SDL_GetError( );
	      
	      if( sdl_error[0] != '\0' )
		{
		  cerr << "ERROR: SDL error '" << sdl_error << "'.\n";
		  SDL_ClearError();
		}
	      
	      resizeScene(Window_Width, Window_Height);
	      frame_counter++;
	      
	      if (SDL_PollEvent(&event))
		done = handleEvent(&event);
	      
	      updateState();
	    }

	  cerr << frame_counter/20.0 << "},";
	}
      cerr << "}\n";
    }
  else
    {
      bool done = false;
      
      while(!done)
	{
	  GLenum gl_error;
	  char* sdl_error;
	  SDL_Event event;
	  
	  /* Check for error conditions. */
	  gl_error = glGetError( );
	  
	  if(gl_error != GL_NO_ERROR)
	    cerr << "ERROR: OpenGL error: '" << gl_error << "': "
		 << gluErrorString(gl_error) << "\n";
	  
	  sdl_error = SDL_GetError( );
	  
	  if( sdl_error[0] != '\0' )
	    {
	      cerr << "ERROR: SDL error '" << sdl_error << "'.\n";
	      SDL_ClearError();
	    }
	  
	  /* Check if there's a pending event. */
	  while (!done)
	    {
	      if (SDL_PollEvent(&event))
		done = handleEvent(&event);
	      
	      updateState();
	    }
	}
    }

  {
    //ofstream out("tiger.smo");
    //out << *scene;
  }

  delete scene;

  cerr << "Quitting SDL... ";
  SDL_Quit();
  cerr << "done.\n";

  return 0;
}

int
main(int argc,
     char **argv)
{
  cout << "Program started.\n";

  runSmoke(argc, argv, 0.0);

  return EXIT_SUCCESS;
}
