/** file: scene.cc
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "scene.h"

#include <GL/glu.h>
#include <GL/glext.h>
#include <cmath>
#include <iostream>



extern int Window_Width, Window_Height;
extern bool fsantialiasing, wireframe;



// Scene::Pixmap

Scene::Pixmap::Pixmap(unsigned int w,
		      unsigned int h)
  : //GLBuffer(false),
    width(w),
    height(h),
    image(w*h*4)
{
  fill(image.begin(), image.end(), 0);
}

void
Scene::Pixmap::copySubImage(iVec r,
			    iVec d,
			    std::vector<GLubyte> &subimage)
{
  std::cerr << "Copying subimage to " << r << " with dimensions " << d << ".\n";
  for (std::vector<GLubyte>::iterator it=subimage.begin(); it!=subimage.end(); it++)
    if (int(*it)!=0) std::cerr << int(*it) << " ";
  for (int y=0; y<d(1); y++)
    {
      std::copy(subimage.begin() + y*d(0),
	        subimage.begin() + d(1) + y*d(0),
	        image.begin() + r(0) + (y+r(1))*width);
    }
  std::cerr << "\n";
}

void
Scene::Pixmap::saveAsPNG(const std::string &filename)
{
  // This needs the encodePNG() function which will be inherited from our
  // glbuffer library
  /*
  vector<GLubyte> data;

  try
    {
      encodePNG(data, width, height, image);
    }
  catch (...)
    {
      cerr << "Failed to generate PNG data.\n";
      throw;
    }

  ofstream str(filename.c_str());

  if (!str)
    {
      cerr << "Failed to open output file `" << filename << "'.\n";
      throw;
    }

  str.write(data.begin(), data.size());
  */
}

void
Scene::Pixmap::makeCurrent()
{
}



// Scene

Scene::Scene()
  : store(new Store),
    root(new Group),
    w(Window_Width),
    h(Window_Height),
    renderaa_texid(0)
{
  if (fsantialiasing)
    {
      glGenTextures(1, &renderaa_texid);
      
      int s=2;
      renderaa_tex.resize(w*h*4);
      
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D, renderaa_texid);
      
      // Set the parameters of the texture
      glTexParameteri(GL_TEXTURE_2D,
		      GL_TEXTURE_WRAP_S,
		      GL_REPEAT);
      glTexParameteri(GL_TEXTURE_2D,
		      GL_TEXTURE_WRAP_T,
		      GL_REPEAT);
      glTexParameteri(GL_TEXTURE_2D,
		      GL_TEXTURE_MAG_FILTER,
		      GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D,
		      GL_TEXTURE_MIN_FILTER,
		      GL_LINEAR_MIPMAP_LINEAR);
      glTexParameteri(GL_TEXTURE_2D,
		      GL_GENERATE_MIPMAP_SGIS,
		      GL_TRUE);
      
      // Create enough mipmap levels in the texture
      for (int i=0; i<s; i++)
	glTexImage2D(GL_TEXTURE_2D,
		     i,
		     GL_RGBA,
		     w>>i,
		     h>>i,
		     0,
		     GL_RGBA,
		     GL_UNSIGNED_BYTE,
		     &*renderaa_tex.begin());
      
      glDisable(GL_TEXTURE_2D);
    }
}

Scene::~Scene()
{
  std::cerr << "Deleting scene:\n";
  std::cerr << "  Deleting antialiased rendering texture.\n";
  glDeleteTextures(1, &renderaa_texid);
  std::cerr << "  Deleting scene tree.\n";
  delete root;
  std::cerr << "  Deleting store.\n";
  delete store;
  std::cerr << "done.\n";
}

const Store &
Scene::getStore() const
{
  return *store;
}

const Group &
Scene::getRoot() const
{
  return *root;
}

Store &
Scene::getStore()
{
  return *store;
}

Group &
Scene::getRoot()
{
  return *root;
}

void
Scene::render(const ViewPort &viewport)
{
  // FIXME: Check for a valid OpenGL context before using it
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  const Bound &b=viewport.getBound();
  gluOrtho2D(b.l(0), b.u(0),
	     b.l(1), b.u(1));
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  if (wireframe)
    {
      glPolygonMode(GL_FRONT, GL_LINE);
      glPolygonMode(GL_BACK, GL_LINE);
      if (fsantialiasing)
	glLineWidth(4);
      else
	glLineWidth(1);
    }
  getRoot().render(viewport);
  /*
  //ViewPort tvp(Bound(b.l + (b.u-b.l)/3., b.l + (b.u-b.l)*2./3.),
  //       viewport.getDelta()/3.);
  getRoot().render(tvp);
  */
  if (wireframe)
    {
      glPolygonMode(GL_FRONT, GL_FILL);
      glPolygonMode(GL_BACK, GL_FILL);
    }
}

void
Scene::renderAA(int width,
		int height,
		const ViewPort &viewport)
{
  int s=2;

  // Calculate the dimensions of the full viewport and the dimensions of the
  // portion which we are currently drawing
  //int sw=w>>s, sh=h>>s;

  // Get the bound of the full viewport
  const Bound &b=viewport.getBound();

  // Calculate the range of indices
  int imax=1<<s, jmax=1<<s;

  // Calculate the index of the current portion
  for (int i=0; i<imax; i++)
    for (int j=0; j<jmax; j++)
      {
	// Calculate the canvas bound corresponding to this portion
	Bound sb(Vec(b.l(0) + (b.u(0) - b.l(0))*i/imax,
		     b.l(1) + (b.u(1) - b.l(1))*j/jmax),
		 Vec(b.l(0) + (b.u(0) - b.l(0))*(i+1)/imax,
		     b.l(1) + (b.u(1) - b.l(1))*(j+1)/jmax));
	
	// Clear the back buffer
	glDrawBuffer(GL_BACK);
	glClear(GL_COLOR_BUFFER_BIT);
	
	// Render the current portion of the scene into the back buffer
	//glEnable(GL_MULTISAMPLE_ARB);
	//glHint(GL_MULTISAMPLE_FILTER_HINT_NV, GL_NICEST);
	render(ViewPort(sb, viewport.getDelta()));///Num(1<<s)));
	//glHint(GL_MULTISAMPLE_FILTER_HINT_NV, GL_FASTEST);
	//glDisable(GL_MULTISAMPLE_ARB);
	glDrawBuffer(GL_FRONT);
	
	// Start using our texture
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, renderaa_texid);

	// Copy the contents of the back buffer into the texture
	glReadBuffer(GL_BACK);
	glCopyTexSubImage2D(GL_TEXTURE_2D,
			    0,
			    0, 0,
			    0, 0,
			    w, h);
	
	// Render the mipmapped texture into the front buffer
	glDrawBuffer(GL_FRONT);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, imax,
		   0, jmax);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glColor4f(1, 1, 1, 1);
	glBegin(GL_QUADS);
	glTexCoord2f(0, 0);
	glVertex2f(  i,   j);
	glTexCoord2f(1, 0);
	glVertex2f(i+1,   j);
	glTexCoord2f(1, 1);
	glVertex2f(i+1, j+1);
	glTexCoord2f(0, 1);
	glVertex2f(  i, j+1);
	glEnd();
	
	glDisable(GL_TEXTURE_2D);
	
	glDrawBuffer(GL_BACK);
      }

  glFinish();
}

void
Scene::renderToPixmap()
{
}

void
Scene::calcGeometryBounds() const
{
  const Store &store=getStore();
  for (Store::GeometryList::const_iterator it=store.geometries.begin();
       it!=store.geometries.end(); it++)
    (*it)->calcBound();
}

const Bound &
Scene::getBound() const
{
  return bound;
}

const Bound &
Scene::calcBounds() const
{
  calcGeometryBounds();
  bound = getRoot().calcBound();
  return bound;
}

void
Scene::save(std::ostream &out) const
{
  getStore().save(out);
  getRoot().save(out, &getStore());
}

/*
void
Scene::saveAsPixmap(const ViewPort &viewport,
		    const std::string &filename,
		    GLContext &ctx)
{
  // Create a pixmap to store the final image and to save it
  cerr << "Creating " << viewport.getWidth() << "x" << viewport.getHeight()
       << " pixmap.\n";
  Pixmap pixmap(viewport.getWidth(), viewport.getHeight());

  // Create a pbuffer with the largest possible dimensions
  // OPTIM: Use smaller pbuffer when pixmap size is much smaller than this
  iVec pbuffer_d(256, 256);
  cerr << "Creating " << pbuffer_d(0) << "x" << pbuffer_d(1) << " pbuffer.\n";
  GLPBuffer *pbuffer=new GLPBuffer(pbuffer_d(0), pbuffer_d(1), &ctx);
  cerr << "Created " << pbuffer->getWidth() << "x"
       << pbuffer->getHeight() << " pbuffer.\n";

  // Make the pbuffer the current rendering context
  cerr << "Making the pbuffer the current rendering context.\n";
  pbuffer->makeCurrent();

  // Calculate the required number of tiles
  iVec n((viewport.getWidth() - 1)/pbuffer->getWidth() + 1,
	 (viewport.getHeight() - 1)/pbuffer->getHeight() + 1);
  cerr << "Using " << n(0) << "x" << n(1) << " tiles.\n";

  // Create a new texture
  cerr << "Creating new texture.\n";
  GLuint tex_id=0;
  glGenTextures(1, &tex_id);

  for (int tx=0; tx<n(0); tx++)
    for (int ty=0; ty<n(1); ty++)
      {
	// Calculate the position of this tile within the pixmap
	iVec
	  subimage_r(tx*pbuffer->getWidth(), ty*pbuffer->getHeight()),
	  subimage_d(pbuffer->getWidth(), pbuffer->getHeight());

	// If the tile hangs off the end then crop it
	if (subimage_r(0) + subimage_d(0) > viewport.getWidth())
	  subimage_d(0) = viewport.getWidth() - subimage_r(0);
	if (subimage_r(1) + subimage_d(1) > viewport.getWidth())
	  subimage_d(1) = viewport.getWidth() - subimage_r(1);

	cerr << "Tile at " << subimage_r << " with dimensions "
	     << subimage_d << ".\n";

	// Create a temporary texture to hold the contents of the pbuffer
	std::vector<GLubyte>
	  subimage(subimage_d(0)*subimage_d(1)*4);

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, tex_id);
	glTexImage2D(GL_TEXTURE_2D,
		     0,
		     GL_RGBA,
		     subimage_d(0),
		     subimage_d(1),
		     0,
		     GL_RGBA,
		     GL_UNSIGNED_BYTE,
		     subimage.begin());

	// Calculate the viewport position in canvas coordinates
	Vec
	  vp_d=viewport.getTopRight() - viewport.getBottomLeft(),
	  l=viewport.getBottomLeft() + Vec(subimage_r(0)*vp_d(0)/viewport.getWidth(),
					   subimage_r(1)*vp_d(1)/viewport.getHeight()),
	  u=viewport.getBottomLeft() + Vec((subimage_r(0)+subimage_d(0))*vp_d(0)/viewport.getWidth(),
					   (subimage_r(1)+subimage_d(1))*vp_d(1)/viewport.getHeight());

	cerr << "Viewport at " << l << ", " << u << ".\n";

	// Render to the pbuffer
	cerr << "Rendering tile...\n";
	ViewPort tile_vp(l, u,
			 pbuffer->getWidth(), pbuffer->getHeight());
	preRender(tile_vp);
	render(tile_vp);

	string subfilename(string("test")+string(1, char('0'+tx))+string(1, char('0'+ty))+string(".png"));
	cerr << "Saving tile to `" << subfilename << "'.\n";
	//pbuffer->saveAsPNG(subfilename);

	// Extract the contents of the pbuffer as a texture
	cerr << "Extracting tile as texture.\n";
	glGetTexImage(GL_TEXTURE_2D,
		      0,
		      GL_RGBA,
		      GL_UNSIGNED_BYTE,
		      subimage.begin());

	// Copy the subimage into the pixmap
	cerr << "Copying texture contents into pixmap.\n";
	pixmap.copySubImage(subimage_r,
			    iVec(pbuffer->getWidth(), pbuffer->getHeight()),
			    subimage);
      }

  // Delete our texture
  glDeleteTextures(1, &tex_id);

  delete pbuffer;

  cerr << "Saving pixmap as PNG.\n";
  pixmap.saveAsPNG(filename);
}
  */

std::ostream &
operator<<(std::ostream &out,
	   const Scene &scene)
{
  scene.save(out);

  return out;
}
