/** file: scene.h
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

class Scene;

#ifndef SCENE_H
#define SCENE_H

#include "autoptr.h"
#include "fill_types.h"
#include "stroke.h"
#include "segments.h"
#include "caches.h"
#include "geometry.h"
#include "editing.h"
#include "tree.h"
#include "store.h"

// We will want to use our glbuffer library for off-screen rendering
//#include <glbuffer++/glbuffer.h>
#include <fstream>
#include <list>
#include <vector>
#include <map>
#include <GL/gl.h>



/** Encapsulation of all the objects on a vector graphics canvas.
 *
 *  Local storage is provided for fill types, line styles and curve segments
 *  used by objects in the scene.
 */
class Scene
{
 private:
  Store *store;

  Group *root;

  mutable Bound bound;

  int w, h;

  // When we start using our glbuffer library we will want to derive our pixmap
  // from the general GLbuffer type to inherit saving to PNG
  class Pixmap// : public GLBuffer
    {
    public:
      Pixmap(unsigned int width,
	     unsigned int height);

      /** Copy the given subimage with dimensions d into the pixmap at
       *  location r. */
      void
	copySubImage(iVec r,
		     iVec d,
		     std::vector<GLubyte> &subimage);

      /** Save the pixmap to the given filename as a PNG file. */
      void
	saveAsPNG(const std::string &filename);

    private:
      int
	width,
	height;

      /** Internal representation. */
      std::vector<GLubyte> image;

      void
	makeCurrent();
    };

  GLuint renderaa_texid;
  std::vector<GLubyte> renderaa_tex;

 public:
  /** Construct an empty scene. */
  Scene();

  /** Destruct the scene in order. */
  ~Scene();

  const Store &
    getStore() const;

  const Group &
    getRoot() const;

  Store &
    getStore();

  Group &
    getRoot();

  /** Render the prerendered objects into the current OpenGL context. */
  void
    render(const ViewPort &viewport);

  void
    renderAA(int width,
	     int height,
	     const ViewPort &viewport);

  /** Render the prerendered objects into the given pixmap. */
  void
    renderToPixmap();

  const Bound &
    getBound() const;

  /** Calculate the bounding areas of all geometries in this scene, */
  void
    calcGeometryBounds() const;

  /** Update the bounds of all geometries and then all elements of the scene
   *  tree. */
  const Bound &
    calcBounds() const;

  /** Export this scene in the internal format. */
  void
    save(std::ostream &out) const;

  /*
  void
    saveAsPixmap(const ViewPort &viewport,
		 const std::string &filename,
		 GLContext &ctx)
  */
};

/** Export the given scene in the internal format. */
std::ostream &
operator<<(std::ostream &out,
	   const Scene &scene);



#endif /* SCENE_H */
