/** file: segments.cc
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "segments.h"

#include "store.h"
#include "utility.h"

#include <cmath>



// Segment

Segment::Segment(const Vec r)
  : endpoint(r)
{
}

Segment::~Segment()
{
}

const Vec &
Segment::getEndPoint() const
{
  return endpoint;
}

const Bound &
Segment::getBound() const
{
  return bound;
}

void
Segment::flatten(const Vec &start,
		 const ViewPort &viewport,
		 std::list<Vec> &contour,
		 Vec &max_deviation) const
{
  Segment::flatten(start, viewport.getDelta(), contour, max_deviation);
}

void
Segment::flatten(const Vec &start,
		 const Vec &delta,
		 std::list<Vec> &contour,
		 Vec &max_deviation) const
{
  contour.push_back(getEndPoint());
}



// LineSegment

LineSegment::LineSegment(const Vec r)
  : Segment(r)
{
}

LineSegment::~LineSegment()
{
}

const Bound &
LineSegment::calcBound(const Vec &s) const
{
  const Vec &e=getEndPoint();

  return bound = Bound(Vec(std::min(s(0), e(0)), std::min(s(1), e(1))),
		       Vec(std::max(s(0), e(0)), std::max(s(1), e(1))));
}

void
LineSegment::save(std::ostream &out,
		  const Store *store) const
{
  const Vec
    &e = getEndPoint();
  out << "Line " << e(0) << " " << e(1);
}



// BezierSegment

BezierSegment::BezierSegment(const Vec c1,
			     const Vec c2,
			     const Vec r)
  : Segment(r),
    child(NULL, NULL),
    controlpoint(c1, c2),
    deviation(Vec(0,0))
{ 
}

BezierSegment::~BezierSegment()
{
  for (int i=0; i<2; i++)
    delete child(i);
}

Vec
BezierSegment::calcDeviation(const Vec start) const
{
  const Vec r = getEndPoint() - start;
  const Num l2 = dot(r, r);
  const Vec
    &c1 = controlpoint(0),
    &c2 = controlpoint(1);
  const Vec
    a = c1-start,
    b = c2-start;
  
  // Special case end point equal to start point
  if (l2 == 0)
    return Vec(std::max(sqrt(a(0)*a(0)), sqrt(b(0)*b(0))),
	       std::max(sqrt(a(0)*a(0)), sqrt(b(0)*b(0))));

  const Num
    x1 = std::abs(a(0)*r(1) - a(1)*r(0)),
    x2 = std::abs(b(0)*r(1) - b(1)*r(0)),
    x = x1 > x2 ? x1 : x2,
    d = x/l2;

  // Return the maximum distance from the approximate line to the exact curve
  // in the direction of the perpendicular to the curve
  return d*Vec(std::abs(r(0)), std::abs(r(1)));
}

const Vec &
BezierSegment::getDeviation() const
{
  return deviation;
}

void
BezierSegment::deCasteljau(const Vec &start,
			   const Num t,
			   TinyVector<Vec, 5> &out) const
{
  const Num t2 = 1-t;

  TinyVector<Vec, 3> b(start*t + controlpoint(0)*t2,
		       controlpoint(0)*t + controlpoint(1)*t2,
		       controlpoint(1)*t + getEndPoint()*t2);

  TinyVector<Vec, 2> bb(b(0)*t + b(1)*t2,
			b(1)*t + b(2)*t2);

  Vec bbb = bb(0)*t + bb(1)*t2;

  out(0) = b(0);
  out(1) = bb(0);
  out(2) = bbb;
  out(3) = bb(1);
  out(4) = b(2);
}

void
BezierSegment::createChildren(const Vec &s) const
{
  const Vec &e = getEndPoint();

  // Bisect this bezier into two Beziers using the de Casteljau
  // contruction
  TinyVector<Vec, 5> out;
  deCasteljau(s, .5, out);

  child(0) = new BezierSegment(out(0), out(1), out(2));
  child(1) = new BezierSegment(out(3), out(4), e);

  child(0)->calcBound(s);
  child(1)->calcBound(child(0)->getEndPoint());
}

void
BezierSegment::flattenChildren(const Vec &start,
			       const ViewPort &viewport,
			       std::list<Vec> &contour,
			       Vec &max_deviation) const
{
  if (!child(0)) createChildren(start);
  child(0)->flatten(start, viewport, contour, max_deviation);
  child(1)->flatten(child(0)->getEndPoint(), viewport, contour, max_deviation);
}

void
BezierSegment::flattenChildren(const Vec &start,
			       const Vec &delta,
			       std::list<Vec> &contour,
			       Vec &max_deviation) const
{
  if (!child(0)) createChildren(start);
  child(0)->flatten(start, delta, contour, max_deviation);
  child(1)->flatten(child(0)->getEndPoint(), delta, contour, max_deviation);
}

void
BezierSegment::flatten(const Vec &start,
		       const ViewPort &viewport,
		       std::list<Vec> &contour,
		       Vec &max_deviation) const
{
  const Vec &delta=viewport.getDelta(), &d=getDeviation();

  if ((d(0)>delta(0) || d(1)>delta(1)) &&
      !viewport.getBound().distinctFrom(getBound()))
    {
      flattenChildren(start, viewport, contour, max_deviation);
      return;
    }

  // Update the maximum deviation for any Beziers inside the viewport
  for (int i=0; i<2; i++)
    if (deviation(i)>max_deviation(i))
      max_deviation(i) = d(i);
  Segment::flatten(start, viewport, contour, max_deviation);
}

void
BezierSegment::flatten(const Vec &start,
		       const Vec &delta,
		       std::list<Vec> &contour,
		       Vec &max_deviation) const
{
  const Vec &d=getDeviation();

  if (d(0)>delta(0) || d(1)>delta(1))
    flattenChildren(start, delta, contour, max_deviation);
  else
    {
      // Update the maximum deviation
      for (int i=0; i<2; i++)
	if (deviation(i)>max_deviation(i))
	  max_deviation(i) = d(i);
      Segment::flatten(start, delta, contour, max_deviation);
    }
}

const Bound &
BezierSegment::calcBound(const Vec &s) const
{
  deviation = calcDeviation(s);
  
  if (child(0))
    {
      child(0)->calcBound(s);
      child(1)->calcBound(child(0)->getEndPoint());
    }

  const Vec
    &c1 = controlpoint(0),
    &c2 = controlpoint(1),
    &e = getEndPoint();

  return bound = Bound(Vec(std::min(std::min(s(0), c1(0)), std::min(c2(0), e(0))),
			   std::min(std::min(s(1), c1(1)), std::min(c2(1), e(1)))),
		       Vec(std::max(std::max(s(0), c1(0)), std::max(c2(0), e(0))),
			   std::max(std::max(s(1), c1(1)), std::max(c2(1), e(1)))));
}

void
BezierSegment::save(std::ostream &out,
		    const Store *store) const
{
  const Vec
    &e = getEndPoint();
  out << "Bezier ";
  for (int i=0; i<2; i++)
    out << controlpoint(i)(0) << " " << controlpoint(i)(1) << " ";
  out << e(0) << " " << e(1);
}
