/** file: segments.h
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

class Segment;
class LineSegment;
class BezierSegment;

#ifndef SEGMENTS_H
#define SEGMENTS_H

class Store;

#include "num.h"
#include "vecmat.h"
#include "viewport.h"
#include "autoptr.h"

#include <list>



/** A path segment.
 *
 *  Paths are trajectories in the plane which can be desribed as lists of
 *  path segments. All path segment types are derived from this base class and
 *  must, therefore, have an end-point and a function to flatten the (possibly
 *  transformed) path to a list of straight lines.
 */
class Segment
{
 private:
  /** All segments finish at the coordinate stored here. */
  Vec endpoint;

 protected:
  /** Bounding area. */
  mutable Bound bound;

 public:
  /** Contruct the base class with the given end point. */
  Segment(const Vec r);

  /** Derived classes may require their own destructors. */
  virtual
    ~Segment();

  /** Retrieve the end point. */
  const Vec &
    getEndPoint() const;

  virtual
    void
    flatten(const Vec &start,
	    const ViewPort &viewport,
	    std::list<Vec> &contour,
	    Vec &max_deviation) const;

  virtual
    void
    flatten(const Vec &start,
	    const Vec &delta,
	    std::list<Vec> &contour,
	    Vec &max_deviation) const;

  /** Retrieve the bottom left hand corner of the bounding box. */
  const Bound &
    getBound() const;

  /** Update the bounding box. */
  virtual
    const Bound &
    calcBound(const Vec &s) const = 0;

  virtual
    void
    save(std::ostream &out,
	 const Store *store) const = 0;
};



/** Straight line path segment.
 */
class LineSegment : public Segment
{
 private:
  Vec endpoint;

 public:
  LineSegment(const Vec r);

  ~LineSegment();

  const Bound &
    calcBound(const Vec &s) const;

  void
    save(std::ostream &out,
	 const Store *store) const;
};



/** Bezier curve path segment.
 *
 *  An implicit bicubic parametric curve described by an implied start point,
 *  two control points and the end-point. Bezier curves are exactly affine
 *  transformed by simply transforming these four points.
 *  FIXME: As segments can be shared between objects we must cache the start
 *  point which was used to generate our children. If it is invalidated then we
 *  must delete and regenerate them.
 *  Also, mutating the start point (i.e. the end-point of the previous segment
 *  in the path) will mean 
 */
class BezierSegment : public Segment
{
 private:
  mutable TinyVector<BezierSegment *, 2> child;

  TinyVector<Vec, 2> controlpoint;

  // FIXME: Does this need to be mutable?
  mutable Vec deviation;

  /** Calculate an upper bound of the maximum deviation of this Bezier curve
   *  from the straight line approximation between the given start point and
   *  the end point. */
  Vec
    calcDeviation(const Vec start) const;

  /** Retrieve the cached deviation of this bezier. */
  const Vec &
    getDeviation() const;

  /** Given the four coordinates of a bezier set the four coordinates of two
   *  beziers which split the original at the given parameter. */
  void
    deCasteljau(const Vec &start,
		const Num t,
		TinyVector<Vec, 5> &out) const;

  /** Bisect this Bezier parametrically to form two smaller, straighter
   *  Beziers. */
  void
    createChildren(const Vec &s) const;

 public:
  BezierSegment(const Vec c1,
		const Vec c2,
		const Vec r);

  ~BezierSegment();

  void
    flattenChildren(const Vec &start,
		    const ViewPort &viewport,
		    std::list<Vec> &contour,
		    Vec &max_deviation) const;

  void
    flattenChildren(const Vec &start,
		    const Vec &delta,
		    std::list<Vec> &contour,
		    Vec &max_deviation) const;

  void
    flatten(const Vec &start,
	    const ViewPort &viewport,
	    std::list<Vec> &contour,
	    Vec &max_deviation) const;

  void
    flatten(const Vec &start,
	    const Vec &delta,
	    std::list<Vec> &contour,
	    Vec &max_deviation) const;

  const Bound &
    calcBound(const Vec &s) const;

  void
    save(std::ostream &out,
	 const Store *store) const;
};



#endif /* SEGMENTS_H */
