/** file: store.cc
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "store.h"
#include <iostream>

Store::Store()
  : font_cache(new FontCache(this))
{
}

Store::~Store()
{
  std::cerr << "  Deleting font cache.\n";
  delete font_cache;
  std::cerr << "  Deleting geometries.\n";
  geometries.clear();
  std::cerr << "  Deleting path segments.\n";
  segments.clear();
  std::cerr << "  Deleting line styles.\n";
  line_styles.clear();
  std::cerr << "  Deleting fill types.\n";
  fill_types.clear();
}

template<class It>
void
Store::saveContainer(It begin,
		     It end,
		     std::ostream &out) const
{
  for (It it=begin; it!=end; it++)
    {
      (*it)->save(out, this);
      out << std::endl;
    }
  out << std::endl;
}

void
Store::saveFillTypes(std::ostream &out) const
{
  saveContainer(fill_types.begin(), fill_types.end(), out);
}

void
Store::saveLineStyles(std::ostream &out) const
{
  saveContainer(line_styles.begin(), line_styles.end(), out);
}

void
Store::saveSegments(std::ostream &out) const
{
  saveContainer(segments.begin(), segments.end(), out);
}

ContourGeometry &
Store::appendContourGeometry()
{
  geometries.push_back(new ContourGeometry);
  return *geometries.back();
}

void
Store::save(std::ostream &out) const
{
  saveFillTypes(out);
  saveLineStyles(out);
  saveSegments(out);
}



// FontCache

FontCache::FontCache(Store *s)
  : store(s)
{
  // Initialise the FreeType library
  if (FT_Init_FreeType(&library))
    {
      std::cerr << "ERROR: Whilst creating FreeType library.\n";
      throw;
    }
}

FontCache::~FontCache()
{
  while (!cache.empty())
    {
      delete cache.begin()->second;
      cache.erase(cache.begin());
    }
}

void
FontCache::loadFont(std::string font_name,
		    std::string file_name)
{
  // Make sure this font name doesn't already exist in the cache
  Cache::iterator it=cache.find(font_name);
  if (it != cache.end())
    {
      std::cerr << "Attempt to load an already loaded font.\n";
      throw;
    }

  // Create a new FreeType font
  cache[font_name] = new CachedFont(store, library, file_name);
}

ContourGeometry &
FontCache::fetchGlyph(std::string font_name,
		      char c)
{
  // Try to find the font in the cache
  Cache::iterator it=cache.find(font_name);
  if (it == cache.end())
    {
      std::cerr << "ERROR: Attempt to fetch glyph from a font which was not loaded.\n";
      throw;
    }

  Entry entry=it->second;
  Store::GeometryList::iterator git=entry->fetchGlyph(c);
  return **git;
}

Vec
FontCache::getGlyphAdvance(std::string font_name,
			   char c) const
{
  Cache::const_iterator it=cache.find(font_name);
  if (it == cache.end())
    {
      std::cerr << "ERROR: Attempt to use unloaded glyph.\n";
      throw;
    }

  Entry entry=it->second;
  return entry->getGlyphAdvance(c);
}

Vec
FontCache::getGlyphAdvance(std::string font_name,
			   char c1,
			   char c2) const
{
  Cache::const_iterator it=cache.find(font_name);
  if (it == cache.end())
    {
      std::cerr << "ERROR: Attempt to use unloaded glyph.\n";
      throw;
    }

  Entry entry=it->second;

  return entry->getGlyphAdvance(c1, c2);
}



// FontCache::CachedFont

FontCache::CachedFont::CachedFont(Store *s,
				  FT_Library &library,
				  std::string file_name)
  : store(s)
{
  if (FT_New_Face(library,
		  file_name.c_str(),
		  0,
		  &face))
    {
      std::cerr << "ERROR: Whilst creating FreeType face.\n";
      throw;
    }
}

void
FontCache::CachedFont::decompose(FT_Outline *outline,
				 ContourGeometry &geo,
				 Num scale)
{
  // Intermediate representation
  std::list<std::list<std::list<Vec> > > l;

  convertFromTrueType(outline, scale, l);

  convertToSegments(geo, l);
}

void
debugOutput(std::list<std::pair<Vec, char> > &t)
{
  typedef std::list<std::pair<Vec, char> >::iterator iterator;
  iterator it=t.begin();

  while (it!=t.end())
    {
      switch (it->second)
	{
	case FT_Curve_Tag_On :
	  std::cout << "  On: ";
	  break;
	case FT_Curve_Tag_Conic :
	  std::cout << "Off2: ";
	  break;
	case FT_Curve_Tag_Cubic :
	  std::cout << "Off3: ";
	  break;
	}
      std::cout << it->first << std::endl;
      it++;
    }
  std::cout << std::endl;
}

void
FontCache::CachedFont::convertFromTrueType(FT_Outline *outline,
					   Num scale,
					   std::list<std::list<std::list<Vec> > > &l)
{
  typedef std::list<std::pair<Vec, char> >::iterator iterator;

  int begin=0, end;
  // Loop through all contours
  for (int contour_i=0; contour_i<outline->n_contours; contour_i++)
    {
      l.push_back(std::list<std::list<Vec> >());
      std::list<std::list<Vec> > &contour=l.back();
      end = outline->contours[contour_i];

      std::list<std::pair<Vec, char> > t;

      // Copy the FT_Outline data structure into an STL container for easier
      // manipulation
      for (int i=begin; i<=end; i++)
	{
	  Vec r=scale*Vec(outline->points[i].x,
			  outline->points[i].y);
	  t.push_back(std::pair<Vec, char>(r, outline->tags[i]));
	}

      //debugOutput(t);

      // Loop through the path segments in this contour, filling in the
      // implicit end-points
      {
	// Move the start point from the beginning of the contour to the end
	// of the contour as it will be the end point of the last segment
	t.insert(t.end(), t.front());
	t.erase(t.begin());

	iterator it=t.begin(), it_1=t.end(), it_2;
	--it_1;
	it_2=it_1;
	--it_2;

	while (it!=t.end())
	  {
	    if (it->second != FT_Curve_Tag_On)
	      {
		if ((it_1->second == FT_Curve_Tag_Conic) ||
		    (it_2->second == FT_Curve_Tag_Cubic &&
		     it_1->second == FT_Curve_Tag_Cubic))
		  {
		    // Make the implicit path segment end explicit by inserting
		    // it after the control points
		    Vec e=.5*(it->first + it_1->first);
		    it_2=it_1;
		    iterator tit=it_1;
		    ++tit;
		    t.insert(tit, std::pair<Vec, char>(e, FT_Curve_Tag_On));
		    ++it_1;
		  }
	      }

	    it_2=it_1;
	    it_1=it;
	    it++;
	  }
      }

      //debugOutput(t);

      {
	iterator it=t.begin();

	contour.push_back(std::list<Vec>());
	std::list<Vec> *segment=&contour.back();

	while (it!=t.end())
	  {
	    segment->push_back(it->first);

	    if (it->second == FT_Curve_Tag_On)
	      {
		it++;
		if (it!=t.end())
		  {
		    contour.push_back(std::list<Vec>());
		    segment=&contour.back();
		  }
	      }
	    else
	      it++;
	  }
      }

      /*
      std::cout << endl;
      for (list<list<Vec> >::iterator it=contour.begin(); it!=contour.end();
	   it++)
	{
	  for (list<Vec>::iterator it2=it->begin(); it2!=it->end(); it2++)
	    std::cout << *it2 << "  ";
	  std::cout << endl;
	}
      */

      begin = end+1;
    }
}

void
FontCache::CachedFont::convertToSegments(ContourGeometry &geo,
					 std::list<std::list<std::list<Vec> > > &l)
{
  if (l.empty()) return;
  if (l.size()==1 && l.front().empty()) return;
  if (l.front().size()==1 && l.front().front().empty()) return;

  // Loop through all contours
  for (std::list<std::list<std::list<Vec> > >::iterator it=l.begin(); it!=l.end(); it++)
    {
      geo.push_back(Contour());
      Contour &contour=geo.back();

      // The start coordinate of the current segment begins with the end point
      // of the last segment
      Vec s=it->back().back();

      convertContourToSegments(*it, contour, s);
    }
}

void
FontCache::CachedFont::convertContourToSegments(std::list<std::list<Vec> > &l,
						Contour &contour,
						Vec &s)
{
  // Loop through all segments in this contour
  for (std::list<std::list<Vec> >::iterator it2=l.begin(); it2!=l.end(); it2++)
    {
      std::list<Vec> &seg=*it2;

      switch (seg.size())
	{
	case 1:
	  {
	    // Push a line segment onto the contour
	    Vec e=seg.back();
	    lines.push_back(LineSegment(e));
	    contour.push_back(&lines.back());
	    break;
	  }
	case 2:
	  {
	    // Convert a conic Bezier segment into a cubic Bezier segment
	    // and push it onto the contour
	    Vec c=seg.front(), e=seg.back();
	    Vec c1=(1/3.)*(s + 2*c), c2=(1/3.)*(2*c + e);
	    beziers.push_back(BezierSegment(c1, c2, e));
	    contour.push_back(&beziers.back());
	    break;
	  }
	case 3:
	  {
	    // Push a cubic Bezier segment onto the contour
	    Vec c1=seg.front(), c2=*(++seg.begin()), e=seg.back();
	    beziers.push_back(BezierSegment(c1, c2, e));
	    contour.push_back(&beziers.back());
	    break;
	  }
	default:
	  {
	    break;
	  }
	}
      
      // Update the start coordinate for the next segment
      s = seg.back();
    }
}

Store::GeometryList::iterator
FontCache::CachedFont::fetchGlyph(char l)
{
  // Get the glyph index of the requested character
  int c=FT_Get_Char_Index(face, l);

  // Try to find the requested glyph in the cache
  Cache::iterator it=cache.find(c);
  if (it != cache.end())
    return it->second;

  // Create a new cache entry containing this glyph

  // Create a new contour geometry for the glyph
  Store::GeometryList &geometries=store->geometries;
  geometries.push_back(new ContourGeometry);
  Store::GeometryList::iterator git=--geometries.end();

  // Enter the contour geometry into the cache
  cache[c] = git;

  if (FT_Load_Glyph(face, c, FT_LOAD_NO_BITMAP | FT_LOAD_NO_SCALE))
    {
      std::cerr << "ERROR: Failed to load glyph `" << c << "'.\n";
      throw;
    }

  FT_GlyphSlot glyph=face->glyph;

  if (glyph->format != ft_glyph_format_outline)
    {
      std::cerr << "ERROR: Not an outline glyph.\n";
      throw;
    }

  // Fill the geometry with the contours of this glyph
  ContourGeometry &geo=**git;
  decompose(&glyph->outline, geo, 1.0/face->units_per_EM);
  geo.calcBound();

  return git;
}

int
FontCache::CachedFont::getCharIndex(char l)
{
  int c=FT_Get_Char_Index(face, l);

  return c;
}

void
FontCache::CachedFont::loadChar(int c)
{
  if (FT_Load_Glyph(face, c, FT_LOAD_NO_BITMAP | FT_LOAD_NO_SCALE))
    {
      std::cerr << "ERROR: Failed to load glyph `" << c << "'.\n";
      throw;
    }
}

Vec
FontCache::CachedFont::getGlyphAdvance(char l)
{
  loadChar(getCharIndex(l));

  FT_Vector advance=face->glyph->advance;

  Num f=1.0/face->units_per_EM;
  return f*Vec(advance.x, advance.y);
}

Vec
FontCache::CachedFont::getGlyphAdvance(char l1,
				       char l2)
{
  int
    c1=getCharIndex(l1),
    c2=getCharIndex(l2);

  loadChar(c2);

  FT_Vector advance=face->glyph->advance, kerning;

  if (FT_Get_Kerning(face, c1, c2, ft_kerning_unscaled, &kerning))
    {
      std::cerr << "ERROR: Failed to get kerning information between `"
		<< c1 << "' and `" << c2 << "'.\n";
      throw;
    }

  Num f=1.0/face->units_per_EM;
  return f*Vec(advance.x + kerning.x, advance.y + kerning.y);
}
