/** file: store.h
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

class Store;
class FontCache;

#ifndef STORE_H
#define STORE_H

#include "stroke.h"
#include "fill_types.h"
#include "segments.h"
#include "geometry.h"

#include <string>
#include <map>
#include <freetype/freetype.h>



class Store
{
 private:
  template<class T>
    class StoreList : public std::list<AutoPtr<T> >
    {
    public:
      T *
	operator[](int i);
    };

  template<class It>
    void
    saveContainer(It begin,
		  It end,
		  std::ostream &out) const;

  void
    saveFillTypes(std::ostream &out) const;

  void
    saveLineStyles(std::ostream &out) const;

  void
    saveSegments(std::ostream &out) const;

 public:
  mutable FontCache *font_cache;

  Store();

  ~Store();

  StrokeStyle::MitreJoin mitrejoin;
  StrokeStyle::RoundJoin roundjoin;
  StrokeStyle::BevelJoin beveljoin;

  StrokeStyle::EndCap noendcap;
  StrokeStyle::RoundEndCap roundendcap;
  StrokeStyle::ButtEndCap buttendcap;

  typedef StoreList<FillType> FillTypeList;
  typedef StoreList<LineStyle> LineStyleList;
  typedef StoreList<Segment> SegmentList;
  typedef StoreList<ContourGeometry> GeometryList;

  FillTypeList fill_types;
  LineStyleList line_styles;
  SegmentList segments;
  GeometryList geometries;

  ContourGeometry &
    appendContourGeometry();

  void
    save(std::ostream &out) const;
};

template<class T>
T *
Store::StoreList<T>::operator[](int i)
{
  // FIXME: Why do we need the "this->" before calls to local functions?
  typename StoreList<T>::iterator it=this->begin();
  int j=0;
  while (j!=i && it!=this->end())
    {
      it++;
      j++;
    }
  if (it==this->end()) throw;
  return *it;
}



/** Allows the use of fonts in a vector graphics scene.
 *
 *  Fonts must first be loaded from a given file into a given font name in
 *  the font cache.  Glyphs from a loaded font may then be requested as
 *  contour geometries in the scene. Glyphs are loaded and converted into
 *  contours as requested.
 */
class FontCache
{
 private:
  /** Cached glyphs from a particular font. */
  class CachedFont
    {
    private:
      typedef Store::GeometryList::iterator Entry;
      typedef std::map<int, Entry> Cache;
      Cache cache;
      
      FT_Face face;
      
      Store *store;
      
      std::list<LineSegment> lines;
      std::list<BezierSegment> beziers;
      
      ContourGeometry &
	createGlyph(char c);
      
      void
	decompose(FT_Outline *outline,
		  ContourGeometry &geo,
		  Num scale);
      
      void
	convertFromTrueType(FT_Outline *outline,
			    Num scale,
			    std::list<std::list<std::list<Vec> > > &l);
      
      void
	convertToSegments(ContourGeometry &geo,
			  std::list<std::list<std::list<Vec> > > &l);
      
      void
	convertContourToSegments(std::list<std::list<Vec> > &l,
				 Contour &contour,
				 Vec &s);
      
    public:
      CachedFont(Store *s,
		 FT_Library &library,
		 std::string file_name);
      
      Store::GeometryList::iterator
	fetchGlyph(char l);
      
      int
	getCharIndex(char l);
      
      void
	loadChar(int c);
      
      Vec
	getGlyphAdvance(char l);
      
      Vec
	getGlyphAdvance(char l1,
			char l2);
    };
  
  typedef CachedFont * Entry;
  typedef std::map<std::string, Entry> Cache;
  Cache cache;
  
  FT_Library library;
  
  Store *store;
  
 public:
  FontCache(Store *s);
  
  ~FontCache();
  
  void
    loadFont(std::string font_name,
	     std::string file_name);
  
  ContourGeometry &
    fetchGlyph(std::string font_name,
	       char c);
  
  Vec
    getGlyphAdvance(std::string font_name,
		    char c) const;
  
  Vec
    getGlyphAdvance(std::string font_name,
		    char c1,
		    char c2) const;
};



#endif /* STORE_H */
