/** file: stroke.cc
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "stroke.h"

#include "store.h"
#include "segments.h"



// StrokeStyle

StrokeStyle::StrokeStyle(Num w,
			 Join &j,
			 EndCap &ec1,
			 EndCap &ec2)
  : width(w),
    join(j),
    endcap0(ec1),
    endcap1(ec2)
{
}

void
StrokeStyle::renderSegment(ViewPort viewport,
			   const Vec &r1,
			   const Vec &r2) const
{
  // Calculate the offset from one end to the other
  const Vec d=r2-r1;

  const Num d2=dot(d, d);

  if (d2==0) return;

  // Calculate a perpendicular unit vector
  const Vec n=Vec(d(1),-d(0))*(width/sqrt(d2));

  // We add vertices in the middle of straight edges to ensure exact numerical
  // precision during scanconversion
  glBegin(GL_POLYGON);
  glVertex(r1);
  glVertex(r1 + n);
  glVertex(r2 + n);
  glVertex(r2);
  glVertex(r2 - n);
  glVertex(r1 - n);
  glEnd();
}

void
StrokeStyle::renderJoin(ViewPort viewport,
			const Vec &r1,
			const Vec &r2,
			const Vec &r3) const
{
  join.render(width, viewport, r1, r2, r3);
}

void
StrokeStyle::renderEndCap(ViewPort viewport,
			  int cap,
			  const Vec &r1,
			  const Vec &r2) const
{
  if (cap==1)
    endcap0.render(width, viewport, r1, r2);
  else
    endcap1.render(width, viewport, r1, r2);
}

const Num &
StrokeStyle::getWidth() const
{
  return width;
}

bool
StrokeStyle::operator<(const StrokeStyle &s) const
{
  if (getWidth() < s.getWidth()) return true;
  if (getWidth() == s.getWidth())
    {
      if (&join < &s.join) return true;
      if (&join == &s.join)
	{
	  if (&endcap0 < &s.endcap0) return true;
	  if (&endcap0 == &s.endcap0)
	    {
	      if (&endcap1 < &s.endcap1) return true;
	      if (&endcap1 == &s.endcap1)
		return false; // Equals
	    }
	}
    }
  return false; // Greater than
}

// StrokeStyle::MitreJoin

void
StrokeStyle::MitreJoin::render(Num width,
			       ViewPort viewport,
			       const Vec &r1,
			       const Vec &r2,
			       const Vec &r3) const
{
  // Calculate the offsets from one end to the other
  const Vec d1=r2-r1, d2=r3-r2;

  // Calculate a perpendicular vectors the width of the line
  Vec
    n1=Vec(d1(1),-d1(0))*(width/sqrt(dot(d1, d1))),
    n2=Vec(d2(1),-d2(0))*(width/sqrt(dot(d2, d2)));

  // Make sure the normal vectors are pointing outwards
  if (n1(0)*n2(1) - n1(1)*n2(0) < 0)
    {
      n1 = -1*n1;
      n2 = -1*n2;
    }

  glBegin(GL_POLYGON);
  glVertex(r2);
  glVertex(r2 + n1);
  glVertex(r2 + n2);
  glEnd();
}

// StrokeStyle::BevelJoin

void
StrokeStyle::BevelJoin::render(Num width,
			       ViewPort viewport,
			       const Vec &r1,
			       const Vec &r2,
			       const Vec &r3) const
{
  // Calculate the offsets from one end to the other
  const Vec d1=r2-r1, d2=r3-r2;

  // Calculate a perpendicular vectors the width of the line
  Vec
    n1=Vec(d1(1),-d1(0))*(width/sqrt(dot(d1, d1))),
    n2=Vec(d2(1),-d2(0))*(width/sqrt(dot(d2, d2)));

  // Make sure the normal vectors are pointing outwards
  if (n1(0)*n2(1) - n1(1)*n2(0) < 0)
    {
      n1 = -1*n1;
      n2 = -1*n2;
    }

  glBegin(GL_POLYGON);
  glVertex(r2);
  glVertex(r2 + n1);
  glVertex(r2 + n2);
  glEnd();
}

// StrokeStyle::RoundJoin

void
StrokeStyle::RoundJoin::render(Num width,
			       ViewPort viewport,
			       const Vec &r1,
			       const Vec &r2,
			       const Vec &r3) const
{
  // OPTIM: Cache the beziers used to stroke rounded joins.

  // Calculate the offsets from one end to the other
  Vec d1=r2-r1, d2=r3-r2;

  Num md1=sqrt(dot(d1, d1)), md2=sqrt(dot(d2, d2));

  // Calculate perpendicular vectors with the length of the line width
  Vec
    n1=Vec(d1(1),-d1(0))*(width/md1),
    n2=Vec(d2(1),-d2(0))*(width/md2);

  int sign=-1;

  // Make sure the normal vectors are pointing outwards
  if (n1(0)*n2(1) - n1(1)*n2(0) < 0)
    {
      sign = 1;
      n1 = -sign*n1;
      n2 = -sign*n2;
    }

  Vec s=d1/md1, e=d2/md2, t=s+e;
  Num
    mt=sqrt(dot(t, t)),
    a=e(1)*s(0) - e(0)*s(1),
    b=e(0)*s(1) - e(1)*s(0);

  Num lambda, mu;

  if (std::abs(mt*a)==0 || std::abs(mt*b)==0)
    lambda=mu=0;
  else
    {
      lambda=4*sign*width*dot(e, t)*(mt - 2)/(3*mt*a);
      mu    =4*sign*width*dot(s, t)*(mt - 2)/(3*mt*b);
    }

  Vec
    p1=r2 + n1,
    p4=r2 + n2,
    p2=p1 + lambda*s,
    p3=p4 + mu*e;
  std::list<Vec> contour;
  Vec max_deviation;
  BezierSegment bez(p2, p3, p4);
  bez.calcBound(p1);
  bez.flatten(p1, viewport, contour, max_deviation);

  glBegin(GL_POLYGON);
  glVertex(r2);
  glVertex(p1);
  for (std::list<Vec>::iterator it=contour.begin(); it!=contour.end(); it++)
    glVertex(*it);
  glEnd();
}

// StrokeStyle::EndCap

void
StrokeStyle::EndCap::render(Num width,
			    ViewPort viewport,
			    const Vec &r1,
			    const Vec &r2) const
{
}

// StrokeStyle::ButtEndCap

void
StrokeStyle::ButtEndCap::render(Num width,
				ViewPort viewport,
				const Vec &r1,
				const Vec &r2) const
{
  // Calculate the offset from one end to the other
  const Vec d=r2-r1;
  const Vec dhat=d*(width/sqrt(dot(d, d)));

  // Calculate a perpendicular unit vector
  const Vec n=Vec(dhat(1),-dhat(0));

  // We add vertices in the middle of straight edges to ensure exact numerical
  // precision during scanconversion
  glBegin(GL_POLYGON);
  glVertex(r1);
  glVertex(r1 + n);
  glVertex(r1 + n - dhat);
  glVertex(r1 - n - dhat);
  glVertex(r1 - n);
  glEnd();
}

// StrokeStyle::RoundEndCap

void
StrokeStyle::RoundEndCap::render(Num width,
				 ViewPort viewport,
				 const Vec &r1,
				 const Vec &r2) const
{
}



std::ostream &
operator<<(std::ostream &out,
	   const StrokeStyle &s)
{
  return out /*<< s.getWidth() << "\n"
	     << s.getJoin().asString() << "\n"
	     << s.getEndCap(0).asString() << "\n"
	     << s.getEndCap(1).asString() << "\n"*/;
}



LineStyle::LineStyle(StrokeStyle s)
  : fill_types(),
    stroke_style(s)
{
}

void
LineStyle::save(std::ostream &out,
		const Store *store) const
{
  out //<< store->fill_types.getIndex(fill) << " "
    << stroke_style;
}
