/** file: tesselate.cc
 *
 *  These functions allow a general polygon to be tesselated in software to
 *  produce a set of non-overlapping triangles. Tesselation is performed using
 *  the latest GLU polygon tesselator implementation in Mesa.
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "tesselate.h"
#include <iostream>

// Global store for temporary vertices
std::list<TinyVector<double, 2> > tess_temps;

GLUtesselator *tobj;

void
error_callback(GLenum err)
{
  std::cerr << "GLU ERROR: " << gluErrorString(err) << std::endl;
  throw;
}

void
combine_callback(GLdouble coords[3],
		 char *vertex_data[4],
		 GLfloat weight[4],
		 GLdouble **data)
{
  typedef TinyVector<double, 2> dVec2;

  // Create a temporary vertex
  tess_temps.push_back(*((dVec2 *)coords));

  // Return it
  *data = (GLdouble *)(&(tess_temps.back()));
}

void
newTesselator(const GLenum winding_rule)
{
  tobj=gluNewTess();

  if (!tobj)
    {
      std::cerr << "ERROR: Failed to allocate GLU tesselator.\n";
      throw;
    }

  gluTessProperty(tobj,
		  GLU_TESS_WINDING_RULE,
		  winding_rule);

  gluTessNormal( tobj, 0.0, 0.0, 1.0 );
  gluTessCallback(tobj, GLU_TESS_BEGIN, (_GLUfuncptr)glBegin);
  gluTessCallback(tobj, GLU_TESS_VERTEX, (_GLUfuncptr)glVertex2dv);
  gluTessCallback(tobj, GLU_TESS_END, (_GLUfuncptr)glEnd);
  gluTessCallback(tobj, GLU_TESS_ERROR, (_GLUfuncptr)error_callback);
  gluTessCallback(tobj, GLU_TESS_COMBINE, (_GLUfuncptr)combine_callback);
}

void
deleteTesselator()
{
  gluDeleteTess(tobj);
}

void
tesselateContour(std::list<TinyVector<double, 3> > &coords,
		 std::list<std::vector<char> > &vertex_data,
		 const std::list<Vec> &contour)
{
  typedef TinyVector<double, 2> dVec2;

  gluTessBeginContour(tobj);

  std::list<Vec> intermediate;
  std::list<Vec>::const_iterator it2=contour.begin();

  while (it2 != contour.end())
    {
      const Vec &r=*it2;
      coords.push_back(TinyVector<double, 3>(r(0), r(1), 0.0));

      vertex_data.push_back(std::vector<char>());

      vertex_data.back().resize(sizeof(dVec2));

      {
	char *array=&(vertex_data.back().front());
	dVec2 &v=*(dVec2 *)(array);
	
	v = dVec2(r(0), r(1));
      }

      gluTessVertex(tobj,
		    (GLdouble *)&(coords.back()),
		    (void *)&(vertex_data.back().front()));

      intermediate.push_back(r);

      it2++;
    }
  
  gluTessEndContour(tobj);
}

void
renderTesselation(const std::list<std::list<Vec> > &polygon,
		  GLenum winding_rule)
{
  newTesselator(winding_rule);

  gluTessBeginPolygon(tobj, NULL);

  std::list<TinyVector<double, 3> > coords;
  std::list<std::vector<char> > vertex_data;

  typedef std::list<std::list<Vec> >::const_iterator const_iterator;

  for (const_iterator it=polygon.begin(); it!=polygon.end(); it++)
    tesselateContour(coords, vertex_data, *it);

  gluTessEndPolygon(tobj);

  deleteTesselator();

  tess_temps.clear();
}
