/** file: test.cc
 *
 *  This is an example of how to use the current Smoke API.
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "test.h"

#include "store.h"

#include <cmath>
#include <cstdlib>
#include <iostream>


// Our global scene. The scene contains everything required to render.
Scene *scene;



void
initFills()
{
  // Shortcut to the list of fill types
  Store &store=scene->getStore();
  Store::FillTypeList &fill_types=store.fill_types;

  // Create some fill types, allocating new storage and putting the pointers
  // on the back of the `fill_types' list.
  fill_types.push_back(new FlatFill(Colour(0, 0, 1, 1)));
  fill_types.push_back(new FlatFill(Colour(1, 1, 1, 1)));
  fill_types.push_back(new FlatFill(Colour(1, 0, 0, .5)));
  fill_types.push_back(new GradientFill(Colour(1, .8, .6, 1),
				   Colour(1, 0, 0, 1),
				   Vec(.4, .2),
				   Vec(.8, .8)));
  fill_types.push_back(new FlatFill(Colour(0, 0, 0, 1)));
}

void
initLineStyles()
{
  // Shortcuts to the lists of fill types and line styles in the scene
  Store &store=scene->getStore();
  Store::FillTypeList &fill_types=store.fill_types;
  Store::LineStyleList &line_styles=store.line_styles;

  // Create some line styles
  // First create a line style with a stroke style and no fills
  line_styles.push_back(new LineStyle(StrokeStyle(.02,
						  store.mitrejoin, 
						  store.buttendcap, 
						  store.buttendcap)));
  // Then add the fills
  line_styles.back()->fill_types.push_back(fill_types[0]);
  
  line_styles.push_back(new LineStyle(StrokeStyle(.01,
						  store.mitrejoin, 
						  store.buttendcap, 
						  store.buttendcap)));
  line_styles.back()->fill_types.push_back(fill_types[1]);
  
  line_styles.push_back(new LineStyle(StrokeStyle(.005,
						  store.mitrejoin, 
						  store.buttendcap, 
						  store.buttendcap)));
  line_styles.back()->fill_types.push_back(fill_types[0]);
  
  line_styles.push_back(new LineStyle(StrokeStyle(.005,
						  store.mitrejoin, 
						  store.buttendcap, 
						  store.buttendcap)));
  line_styles.back()->fill_types.push_back(fill_types[2]);
}

void
initObjects1()
{
  // Create the figure 8

  // Shortcuts
  Store &store=scene->getStore();
  Store::LineStyleList &line_styles=store.line_styles;
  Store::SegmentList &segments=store.segments;

  // Create a new contour object which uses our contour geometry with the
  // given winding rule
  ContourObject
    *obj=new ContourObject(&store, GLU_TESS_WINDING_ODD);

  // No interior fills for this object

  // Add line styles to the object
  obj->line_styles.push_back(line_styles[0]);
  obj->line_styles.push_back(line_styles[1]);
  obj->line_styles.push_back(line_styles[2]);

  // Add this object to the end of the group at the root of the scene tree
  scene->getRoot().push_back(obj);

  // Get a reference to the geometry of the contour object (acts as an STL
  // list of Contours)
  ContourGeometry &geo=obj->getGeometry();

  // Create a new contour in the contour geometry and get a reference to it
  geo.push_back(Contour());
  Contour &contour=geo.back();

  // Create new curve segments in the scene and then push pointers to these
  // onto the end of the current contour
  segments.push_back(new BezierSegment(Vec(2, .1),
				       Vec(-1, .9),
				       Vec(.5, .9)));
  contour.push_back(segments.back());

  segments.push_back(new BezierSegment(Vec(2, .9),
				       Vec(-1, .1),
				       Vec(.5, .1)));
  contour.push_back(segments.back());

  Mat m=translationMatrix(Vec(.5,.5))*rotationMatrix(.5*M_PI)*translationMatrix(Vec(-.5,-.5));
  Transform *trans=new Transform(m);
  trans->push_back(new ContourObject(*obj));
  scene->getRoot().push_back(trans);
}

void
initObjects2()
{
  Store &store=scene->getStore();

  // Object
  ContourObject
    *obj=new ContourObject(&store, GLU_TESS_WINDING_ODD);

  Mat m=translationMatrix(Vec(-.18,0));
  Transform *trans=new Transform(m);
  trans->push_back(obj);
  scene->getRoot().push_back(trans);

  // Interior fills
  Store::FillTypeList &fill_types=store.fill_types;
  obj->fill_types.push_back(fill_types[3]);

  // Stroke line styles
  Store::LineStyleList &line_styles=store.line_styles;
  obj->line_styles.push_back(line_styles[3]);

  ContourGeometry &geo=obj->getGeometry();

  // Create a new contour in this ContourObject and get a reference to it so
  // that we can build it up from curve segments
  geo.push_back(Contour());
  Contour &contour=geo.back();

  // Push bezier curve segments onto the end of this contour in the
  // ContourObject
  Store::SegmentList &segments=store.segments;
  segments.push_back(new BezierSegment(Vec(.4, .2),
				       Vec(.5, .45),
				       Vec(.45, .6)));
  contour.push_back(segments.back());
  segments.push_back(new BezierSegment(Vec(.4, .6),
				       Vec(.35, .6),
				       Vec(.2, .55)));
  contour.push_back(segments.back());
  segments.push_back(new BezierSegment(Vec(.22, .62),
				       Vec(.25, .62),
				       Vec(.25, .7)));
  contour.push_back(segments.back());
  segments.push_back(new BezierSegment(Vec(.5, .7),
				       Vec(.6, .7),
				       Vec(.8, .85)));
  contour.push_back(segments.back());
  segments.push_back(new BezierSegment(Vec(.7, .7),
				       Vec(.65, .6),
				       Vec(.52, .6)));
  contour.push_back(segments.back());
  segments.push_back(new BezierSegment(Vec(.5, .3),
				       Vec(.42, .1),
				       Vec(.25, .4)));
  contour.push_back(segments.back());
}

void
initObjects3()
{
  Store &store=scene->getStore();
  store.font_cache->loadFont("Fontname", "Arial.ttf");

  {
    TextObject *obj=new TextObject(&store,
				   .13,
				   Vec(.35, .45),
				   "Fontname",
				   "on Harrop!");
    obj->fill_types.push_back(store.fill_types[3]);
    obj->line_styles.push_back(store.line_styles[3]);
    scene->getRoot().push_back(obj);
  }

  {
    TextObject *obj=new TextObject(&store,
				   1,
				   Vec(0, 0),
				   "Fontname",
				   "Smoke kicks ass");
    obj->fill_types.push_back(store.fill_types[2]);
    Mat m=translationMatrix(Vec(.9426, .4699))*scaleMatrix(.001);
    Transform *trans=new Transform(m);
    trans->push_back(new TextObject(*obj));
    scene->getRoot().push_back(trans);
    //scene->getRoot().push_back(obj);
  }
}

void
initObjects()
{
  std::cerr << "Building objects for the scene.\n";
  initObjects1();
  initObjects2();
  initObjects3();
}

void
initScene()
{
  scene = new Scene;

  // Initialise the fill types first as these do not reference anything else
  std::cerr << "Creating fill types.\n";
  initFills();

  // Then initialise the line styles which reference the fill types
  std::cerr << "Creating line styles.\n";
  initLineStyles();

  // Create the objects, placing them in the object group which we created at
  // the root of the scene tree
  initObjects();

  std::cerr << "Initialising bounding boxes.\n";
  scene->calcBounds();
}

void
finishScene()
{
  delete scene;
}
