/** file: tree.cc
 *
 *  A scene tree is a retained mode representation of a vector graphics scene,
 *  stored as a directed acyclic graph. Nodes in the scene tree are represented
 *  by the classes here.
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "tree.h"

#include "store.h"
#include <iostream>



extern bool edge_antialiasing, randomised_viewport;



// Element

Element::Element()
{
}

Element::~Element()
{
}

Bound &
Element::getMutableBound() const
{
  return bound;
}

void
Element::setParent(Element *p)
{
  if (parent)
    {
      std::cerr << "ERROR: Attempt to insert an object which is already in the scene tree.\n";
      throw;
    }
  parent = p;
}

const Element *
Element::getParent() const
{
  return parent;
}

const Bound &
Element::getBound() const
{
  return bound;
}



// ContourObject

ContourObject::ContourObject(ContourGeometry &geo,
			     GLenum wr)
  : geometry(geo),
    winding_rule(wr)
{
}

ContourObject::ContourObject(Store *store,
			     GLenum wr)
  : geometry(store->appendContourGeometry()),
    winding_rule(wr)
{
}

ContourObject::~ContourObject()
{
}

ContourGeometry &
ContourObject::getGeometry()
{
  return geometry;
}

const ContourGeometry &
ContourObject::getGeometry() const
{
  return geometry;
}

ContourObject::CachedPartialTesselation::Entry::Entry()
  : dl(0),
    deviation(0,0),
    region()
{
}

ContourObject::CachedPartialTesselation::Entry::Entry(GLDisplayList *l,
						      Vec d,
						      Bound r)
  : dl(l),
    deviation(d),
    region(r)
{
}

ContourObject::CachedPartialTesselation::Entry
ContourObject::renderNewPartialInterior(GLenum winding_rule,
					ViewPort viewport) const
{
  // Create and fill a new display list, enter it into the local
  // tesselation cache and list of automatically deleted display lists
  GLDisplayList &dl=*new GLDisplayList();
  dl.begin();
  Vec deviation = getGeometry().renderInterior(winding_rule, viewport);
  dl.end();
  tesselations.display_lists.push_back(&dl);
  return CachedPartialTesselation::Entry(&dl, deviation, viewport.getBound());
}

void
ContourObject::renderPartialInterior(GLenum winding_rule,
				     ViewPort viewport) const
{
  typedef CachedPartialTesselation::Interiors::iterator iterator;
  iterator it=tesselations.interiors.find(winding_rule);

  // The current cache entry is valid if:
  //  a) there is a suitable cache entry
  //  b) the intersection with the viewport has not expanded
  //  c) the cached deviation is still smaller than the current delta
  if (it!=tesselations.interiors.end())
    {
      CachedPartialTesselation::Entry entry=it->second;
      if (entry.region.covers(intersection(viewport.getBound(), getBound())))
	{
	  const Vec
	    &cached_deviation=entry.deviation,
	    &delta=viewport.getDelta();
	  if (cached_deviation(0) <= delta(0) &&
	      cached_deviation(1) <= delta(1))
	    if(cached_deviation(0) > .1*delta(0) &&
	       cached_deviation(1) > .1*delta(1))
	      {
		entry.dl->exec();
		return;
	      }
	}

      // Delete the existing, invalid cached tesselation ready for regeneration
      tesselations.interiors.erase(it);
    }

  // Create a new region which is slightly bigger than the requested viewport
  // to increase the chances that it will be valid next time
  const Bound &b=viewport.getBound();
  Vec c=.5*(b.l + b.u), r=b.u - c;
  Num s=2, dd=.25;
  if (randomised_viewport)
    {
      s+=.25*(Num(rand())/RAND_MAX);
      dd+=.1*(Num(rand())/RAND_MAX);
    }
  ViewPort ext_viewport(Bound(c - s*r, c + s*r), dd*viewport.getDelta());

  tesselations.interiors[winding_rule]=renderNewPartialInterior(winding_rule,
								ext_viewport);
}

void
ContourObject::renderInterior(ViewPort viewport,
			      bool culling) const
{
  bool is_opaque=true;

  for (FillTypes::const_iterator it=fill_types.begin();
       it!=fill_types.end(); it++)
    {
      FillType &fill=**it;

      // If the fill type is definitely opaque then we don't need blending
      // or depth buffering
      bool will_be_opaque=fill.isOpaque();
      if (will_be_opaque && !edge_antialiasing)
	glDisable(GL_BLEND);
      else
	{
	  glEnable(GL_BLEND);
	  if (edge_antialiasing)
	    glBlendFunc(GL_SRC_ALPHA_SATURATE, GL_ONE);
	  else
	    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
      
      is_opaque=will_be_opaque;

      fill.set(viewport);
      if (culling)
	renderPartialInterior(winding_rule, viewport);
      else
	getGeometry().renderInterior(winding_rule, viewport.getDelta());
      fill.unset();
    }

  if (!edge_antialiasing)
    glDisable(GL_BLEND);
}

ContourObject::CachedPartialTesselation::Entry
ContourObject::renderNewPartialStrokes(const StrokeStyle &style,
				       ViewPort viewport) const
{
  // Create and fill a new display list, enter it into the local
  // tesselation cache and list of automatically deleted display lists
  GLDisplayList &dl=*new GLDisplayList();
  dl.begin();

  // Use a fake viewport area which takes the maximum stroke width into
  // account
  Num w=getMaxStrokeWidth();
  Vec r(w, w);
  ViewPort fake_viewport(Bound(viewport.getBound().l-r,
			       viewport.getBound().u+r),
			 viewport.getDelta());
  Vec deviation = getGeometry().renderStrokes(style, fake_viewport);

  dl.end();
  tesselations.display_lists.push_back(&dl);
  return CachedPartialTesselation::Entry(&dl, deviation, viewport.getBound());
}

void
ContourObject::renderPartialStrokes(const StrokeStyle &style,
				    ViewPort viewport) const
{
  // Calculate the intersection between the viewport and our bounding area
  Bound intrsct=intersection(viewport.getBound(), getBound());

  typedef CachedPartialTesselation::Strokes::iterator iterator;
  iterator it=tesselations.strokes.find(style);

  // The current cache entry is valid if:
  //  b) the intersection with the viewport has not expanded
  //  c) the cached deviation is still smaller than the current delta
  //  d) there is a suitable cache entry
  if (it!=tesselations.strokes.end())
    {
      CachedPartialTesselation::Entry entry=it->second;
      if (entry.region.covers(intrsct))
	{
	  const Vec
	    &cached_deviation=entry.deviation,
	    &delta=viewport.getDelta();
	  if (cached_deviation(0) <= delta(0) &&
	      cached_deviation(1) <= delta(1) &&
	      cached_deviation(0) > .1*delta(0) &&
	      cached_deviation(1) > .1*delta(1))
	    {
	      entry.dl->exec();
	      return;
	    }
	}

      // Delete the existing, invalid cached tesselation ready for regeneration
      tesselations.strokes.erase(it);
    }

  // Create a new region which is slightly bigger than the requested viewport
  // to increase the chances that it will be valid next time
  const Bound &b=viewport.getBound();
  Vec c=.5*(b.l + b.u), r=b.u - c;
  Num s=2, dd=.25;
  if (randomised_viewport)
    {
      s+=.5*(Num(rand())/RAND_MAX);
      dd+=.1*(Num(rand())/RAND_MAX);
    }
  ViewPort ext_viewport(Bound(c - s*r, c + s*r), dd*viewport.getDelta());

  tesselations.strokes[style] = renderNewPartialStrokes(style, ext_viewport);
}

void
ContourObject::renderStrokes(ViewPort viewport,
			     bool culling) const
{
  bool is_opaque=true;

  Vec delta=viewport.getDelta();
  for (LineStyles::const_iterator it=line_styles.begin();
       it!=line_styles.end(); it++)
    {
      LineStyle &line_style=**it;
      StrokeStyle stroke_style=line_style.stroke_style;
      FillTypes &fills=line_style.fill_types;
      for (FillTypes::const_iterator it=fills.begin(); it!=fills.end(); it++)
	{
	  FillType &fill=**it;

	  // If the fill type is definitely opaque then we don't need blending
	  // or depth buffering
	  bool will_be_opaque=fill.isOpaque();
	  if (will_be_opaque && !edge_antialiasing)
	    {
	      glDisable(GL_DEPTH_TEST);
	      glDisable(GL_BLEND);
	    }
	  else
	    {
	      glEnable(GL_DEPTH_TEST);
	      glDepthFunc(GL_LESS);
	      glEnable(GL_BLEND);
	      if (edge_antialiasing)
		glBlendFunc(GL_SRC_ALPHA_SATURATE, GL_ONE);
	      else
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	    }
	  
	  is_opaque=will_be_opaque;

	  if (!is_opaque)
	    glClear(GL_DEPTH_BUFFER_BIT);

	  fill.set(viewport);
	  if (culling)
	    renderPartialStrokes(stroke_style, viewport);
	  else
	    getGeometry().renderStrokes(stroke_style, delta);
	  fill.unset();
	}
    }

  if (!is_opaque)
    {
      // Disable depth testing
      glDisable(GL_DEPTH_TEST);
      if (!edge_antialiasing)
	glDisable(GL_BLEND);
    }
}

void
ContourObject::render(ViewPort viewport) const
{
  const Bound &vpb=viewport.getBound(), &b=getBound();
  if (!vpb.distinctFrom(b))
    {
      const Vec &vpd=viewport.getDelta();
      if ((b.u(0) - b.l(0)) > vpd(0) &&
	  (b.u(1) - b.l(1)) > vpd(1))
	{
	  // Use the local tesselation cache and culled flattening when the
	  // bounding area of this object is considerably larger than the
	  // viewport
	  bool culling=((b.u(0)-b.l(0)) > 16*(vpb.u(0)-vpb.l(0)) ||
			(b.u(1)-b.l(1)) > 16*(vpb.u(1)-vpb.l(1)));

	  renderInterior(viewport, culling);
	  renderStrokes(viewport, culling);
	}
    }
}

Num
ContourObject::getMaxStrokeWidth() const
{
  Num w=0;
  for (LineStyles::const_iterator it=line_styles.begin();
       it!=line_styles.end(); it++)
    {
      LineStyle &ls=**it;
      Num lw=ls.stroke_style.getWidth();
      if (lw > w) w = lw;
    }
  return w;
}

Bound
ContourObject::calcBound() const
{
  // FIXME: When mitred joins are implemented they can extend beyond the stroke
  // width to a controllable maximum.
  Bound &bound=getMutableBound();
  bound = getGeometry().getBound();
  Num w=getMaxStrokeWidth();
  Vec r=Vec(w, w);
  bound.l -= r;
  bound.u += r;
  return bound;
}

void
ContourObject::save(std::ostream &out,
		    const Store *store) const
{
  out << "ContourObject";
}



// Group

Group::Group()
{
}

Group::~Group()
{
  clear();
}

void
Group::render(ViewPort vp) const
{
  for (const_iterator it=begin(); it!=end(); it++)
    (*it)->render(vp);
}

Bound
Group::calcBound() const
{
  Bound &bound=getMutableBound();
  if (empty())
    return bound = Bound();
  const_iterator it=begin();
  {
    bound = (*it)->calcBound();
    it++;
  }
  for (; it!=end(); it++)
    bound = Bound(bound, (*it)->calcBound());
  return bound;
}

void
Group::saveElements(std::ostream &out,
		    const Store *store) const
{
  for (const_iterator it=begin(); it!=end(); it++)
    {
      (*it)->save(out, store);
      out << std::endl;
    }
  out << std::endl;
}

void
Group::save(std::ostream &out,
	    const Store *store) const
{
  out << "Group\n";
  saveElements(out, store);
}



// Transform

Transform::Transform(Mat nm)
  : m(nm),
    im(inverse(m))
{
  convertToGLMatrix(m);
}

Transform::~Transform()
{
}

void
Transform::convertToGLMatrix(Mat m)
{
  // Initialise the matrix to the identity
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      if (i==j)
	glm[i+4*j] = 1;
      else
	glm[i+4*j] = 0;

  // Copy the 2x2 transformation submatrix
  for (int i=0; i<2; i++)
    for (int j=0; j<2; j++)
      glm[i+4*j] = m(i,j);

  // Copy the translation 2-vector
  for (int i=0; i<2; i++)
    glm[i+4*3] = m(i,2);
}

void
Transform::render(ViewPort vp) const
{
  glPushMatrix();
  glMultMatrixf(glm);
  Group::render(im*vp);
  glPopMatrix();
}

Bound
Transform::calcBound() const
{
  Bound &bound=getMutableBound();
  if (empty())
    return bound = Bound();
  const_iterator it=begin();
  {
    bound = (*it)->calcBound();
    it++;
  }
  for (; it!=end(); it++)
    bound = Bound(bound, m*((*it)->calcBound()));
  return bound;
}

void
Transform::save(std::ostream &out,
		const Store *store) const
{
  out << "Transform\n";
  for (int i=0; i<3; i++)
    for (int j=0; j<2; j++)
      out << m(i, j) << "\n";
  Group::saveElements(out, store);
}



// TextObject

TextObject::TextObject(Store *store,
		       Num s,
		       Vec r,
		       std::string font,
		       std::string t)
  : text(t),
    font_name(font),
    scale(s),
    pos(r),
    fill_types(*this),
    line_styles(*this)
{
  typedef std::string::iterator iterator;
  Vec pen=pos;

  for (iterator it=text.begin(); it!=text.end(); it++)
    {
      ContourGeometry &geo=store->font_cache->fetchGlyph(font, *it);
      ContourObject *obj=new ContourObject(geo, GLU_TESS_WINDING_ODD);

      Mat m=translationMatrix(pen)*scaleMatrix(scale);
      Transform *trans=new Transform(m);

      trans->push_back(obj);
      Group::push_back(trans);

      Vec advance;

      if (it==text.begin())
	advance=store->font_cache->getGlyphAdvance(font, *it);
      else
	{
	  iterator oit(it);
	  --oit;
	  advance=store->font_cache->getGlyphAdvance(font, *oit, *it);
	}

      pen += advance*scale;
    }
}

void
TextObject::render(ViewPort vp) const
{
  for (const_iterator it=begin(); it!=end(); it++)
    (*it)->render(vp);
}

Bound
TextObject::calcBound() const
{
  Bound &bound=getMutableBound();
  if (empty())
    return bound = Bound();
  const_iterator it=begin();
  {
    bound = (*it)->calcBound();
    it++;
  }
  for (; it!=end(); it++)
    bound = Bound(bound, (*it)->calcBound());
  return bound;
}

void
TextObject::save(std::ostream &out,
		 const Store *store) const
{
  out << "TextObject\n"
      << font_name << "\n"
      << text.size() << "\n"
      << text << "\n";
}



// TextObject::FillTypes

TextObject::FillTypes::FillTypes(TextObject &o)
  : object(o)
{
}

void
TextObject::FillTypes::push_back(FillType *f)
{
  for (TextObject::iterator it=object.begin(); it!=object.end(); it++)
    {
      Transform *trans=static_cast<Transform *>(&**it);
      ContourObject *obj=static_cast<ContourObject *>(&*trans->front());
      obj->fill_types.push_back(f);
    }
}



// TextObject::LineStyles

TextObject::LineStyles::LineStyles(TextObject &o)
  : object(o)
{
}

void
TextObject::LineStyles::push_back(LineStyle * const &ls)
{
  for (TextObject::iterator it=object.begin(); it!=object.end(); it++)
    {
      Transform *trans=static_cast<Transform *>(&**it);
      ContourObject *obj=static_cast<ContourObject *>(&*trans->front());
      obj->line_styles.push_back(ls);
    }
}
