/** file: tree.h
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

class Element;
class ContourObject;
class Group;
class Transform;
class TextObject;

#ifndef TREE_H
#define TREE_H

class Store;

#include "fill_types.h"
#include "stroke.h"
#include "gldisplaylist.h"
#include "caches.h"
#include "autoptr.h"

#include <string>
#include <list>
#include <map>



/** Abstract base class representing any element in the scene tree. */
class Element
{
 private:
  /** Bounding area of all elements below this one in the scene tree. */
  mutable Bound bound;

  /** Pointer to the element immediately above this one or NULL if this object
   *  is the root of the scene tree.
   *
   *  The onus is on the parent object to set this pointer in its children as
   *  they are inserted.
   */
  Element *parent;

 protected:
  /** Get a mutable reference to the bounding area in order to update it from
   *  a derived class. */
  Bound &
    getMutableBound() const;

  /** Set the parent of this element, objecting if a parent already exists.
   *
   *  This provides a mechanism to ensure that all objects in the scene tree
   *  are unique.
   */
  void
    setParent(Element *);

 public:
  Element();
  
  /** As an abstract base class, we must implement a virtual destructor. */
  virtual
    ~Element();
  
  /** Render this portion of the scene tree. */
  virtual
    void
    render(ViewPort vp) const = 0;

  /** Get a const reference to the pointer to the parent of this element. */
  const Element *
    getParent() const;
  
  /** Get a constant reference to the bounding area of this element. */
  const Bound &
    getBound() const;
  
  /** Update the bounds of this portion of the scene tree. */
  virtual
    Bound
    calcBound() const = 0;
 
  /** Export this element to the given stream. */ 
  virtual
    void
    save(std::ostream &out,
	 const Store *store) const = 0;
};



/** A leaf in the scene tree. */
class ContourObject : public Element
{
 private:
  /** Geometry of this object. */
  ContourGeometry &geometry;
  
  class CachedPartialTesselation
    {
    public:
      class Entry
	{
	public:
	  Entry();
	  
	  Entry(GLDisplayList *l,
		Vec d,
		Bound r);
	  
	  GLDisplayList *dl;
	  
	  Vec deviation;
	  
	  /** Region of the geometry for which this cache is valid. */
	  Bound region;
	};
      
      typedef std::map<GLenum, Entry> Interiors;
      typedef std::map<StrokeStyle, Entry> Strokes;
      
      Interiors interiors;
      Strokes strokes;
      
      /** Automatically deleted store of display lists used by the
       *  cache. */
      std::list<AutoPtr<GLDisplayList> > display_lists;
    };
  
  mutable CachedPartialTesselation tesselations;
  
 public:
  /** Construct with the given geometry. */
  ContourObject(ContourGeometry &geo,
		GLenum wr);
  
  /** Create a geometry in the scene. */
  ContourObject(Store *store,
		GLenum wr);
  
  virtual
    ~ContourObject();
  
  /** Winding rule to be applied to the interior. */
  GLenum winding_rule;
  
  typedef std::list<FillType *> FillTypes;
  typedef std::list<LineStyle *> LineStyles;
  
  /** Fill types to be applied to the interior. */
  FillTypes fill_types;
  
  /** Line styles used to stroke the paths. */
  LineStyles line_styles;
  
  ContourGeometry &
    getGeometry();
  
  const ContourGeometry &
    getGeometry() const;
  
  CachedPartialTesselation::Entry
    renderNewPartialInterior(GLenum winding_rule,
			     ViewPort vp) const;
  
  void
    renderPartialInterior(GLenum winding_rule,
			  ViewPort vp) const;
  
  /** Render the interior in each fill type, possibly using the whole
   *  geometry to eliminate the need for culling. */
  void
    renderInterior(ViewPort vp,
		   bool culling) const;
  
  CachedPartialTesselation::Entry
    renderNewPartialStrokes(const StrokeStyle &style,
			    ViewPort vp) const;
  
  void
    renderPartialStrokes(const StrokeStyle &style,
			 ViewPort vp) const;
  
  /** Stroke the contours with the stroke style of each line style and
   *  render each result in each fill type for that line style, possibly
   *  using the whole geometry to eliminate the need for culling. */
  void
    renderStrokes(ViewPort vp,
		  bool culling) const;
  
  /** Render the interior and strokes of this object. */
  void
    render(ViewPort vp) const;
  
  Num
    getMaxStrokeWidth() const;
  
  /** Update the bound of this object. */
  virtual
    Bound
    calcBound() const;
  
  void
    save(std::ostream &out,
	 const Store *store) const;
  
  void
    clearCache() const;
};



/** A node in the scene tree.
 *
 *  Groups of objects and transormations appear as nodes in the scene tree.
 */
class Group : public Element, public std::list<AutoPtr<Element> >
{
 public:
  Group();
  
  virtual
    ~Group();
  
  virtual
    void
    render(ViewPort vp) const;
  
  /** Update the bound of this object. */
  virtual
    Bound
    calcBound() const;
  
  void
    saveElements(std::ostream &out,
		 const Store *store) const;
  
  virtual
    void
    save(std::ostream &out,
	 const Store *store) const;
};



class Transform : public Group
{
 private:
  /** Inverse of the transformation matrix. */
  Mat m, im;
  
  /** OpenGL compatible version of the transformation matrix. */
  GLfloat glm[16];
  
  /** Convert the given matrix into OpenGL form and store it in `glm'. */
  void
    convertToGLMatrix(Mat m);
  
 public:
  Transform(Mat nm);
  
  virtual
    ~Transform();
  
  void
    render(ViewPort vp) const;
  
  Bound
    calcBound() const;
  
  void
    save(std::ostream &out,
	 const Store *store) const;
};



class TextObject : public Group
{
 private:
  std::string text, font_name;
  
  Num scale;
  
  Vec pos;
  
 public:
  class FillTypes : public std::list<FillType *>
    {
    private:
      TextObject &object;
      
    public:
      FillTypes(TextObject &o);
      
      void
	push_back(FillType *f);
    };
  
  class LineStyles : public std::list<LineStyle *>
    {
    private:
      TextObject &object;
      
    public:
      LineStyles(TextObject &o);
      
      void
	push_back(LineStyle * const &ls);
    };
  
  friend class FillTypes;
  friend class LineStyles;
  
  FillTypes fill_types;
  LineStyles line_styles;
  
  TextObject(Store *store,
	     Num s,
	     Vec r,
	     std::string font,
	     std::string str);
  
  void
    render(ViewPort vp) const;
  
  Bound
    calcBound() const;
  
  virtual
    void
    save(std::ostream &out,
	 const Store *store) const;
};



#endif /* TREE_H */
