/** vecmat.cc
 *
 *  Vector and matrix implementations for 2D vector graphics. The TinyVector
 *  implementation mimics the Blitz++ library but without using template
 *  metaprogramming. This greatly reduces compile-time.
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include "vecmat.h"

#include <GL/gl.h>
#include <cmath>

Vec
operator*(const Mat a,
	  const Vec b)
{
  return Vec(a(0, 0)*b(0) + a(0, 1)*b(1) + a(0, 2),
	     a(1, 0)*b(0) + a(1, 1)*b(1) + a(1, 2));
}

void
getModelView(Mat &m)
{
  GLfloat n[16];
  glGetFloatv(GL_MODELVIEW_MATRIX, n);
  for (int i=0; i<2; i++)
    for (int j=0; j<3; j++)
      m(i, j) = n[i + 4*j];
}

Mat
inverse(const Mat m)
{
  Mat im;

  // Set the transformation elements
  im(0,0) = m(1,1);
  im(0,1) = -m(0,1);
  im(1,0) = -m(1,0);
  im(1,1) = m(0,0);

  // Set the translation elements
  im(0,2) = m(0,1)*m(1,2) - m(1,1)*m(0,2);
  im(1,2) = m(1,0)*m(0,2) - m(0,0)*m(1,2);

  // Divide by the determinant
  {
    Num id=1/(m(0,0)*m(1,1) - m(0,1)*m(1,0));
    for (int i=0; i<2; i++)
      for (int j=0; j<3; j++)
	im(i,j) *= id;
  }

  return im;
}

Mat
rotationMatrix(const Num a)
{
  Mat m;
  Num c=std::cos(a), s=std::sin(a);

  m(0,0) = c;
  m(0,1) = -s;
  m(0,2) = 0;
  m(1,0) = s;
  m(1,1) = c;
  m(1,2) = 0;

  return m;
}

Mat
scaleMatrix(const Num a)
{
  Mat m;

  m(0,0) = a;
  m(0,1) = 0;
  m(0,2) = 0;
  m(1,0) = 0;
  m(1,1) = a;
  m(1,2) = 0;

  return m;
}

Mat
translationMatrix(const Vec r)
{
  Mat m;

  m(0,0) = 1;
  m(0,1) = 0;
  m(0,2) = r(0);
  m(1,0) = 0;
  m(1,1) = 1;
  m(1,2) = r(1);

  return m;
}

Num
length2(Vec r)
{
  return dot(r, r);
}

Num
length(Vec r)
{
  return std::sqrt(length2(r));
}

Mat
operator*(const Mat m1,
	  const Mat m2)
{
  Mat m;

  for (int i=0; i<2; i++)
    for (int j=0; j<3; j++)
      {
	m(i,j)=0;
	for (int k=0; k<2; k++)
	  m(i,j) += m1(i,k)*m2(k,j);
	if (j==2)
	  m(i,j) += m1(i,2);
      }

  return m;
}

std::ostream &
operator<<(std::ostream &out,
	   Mat m)
{
  for (int i=0; i<5; i++)
    out << m(i/3, i%3) << " ";
  out << m(2, 3);

  return out;
}

std::istream &
operator>>(std::istream &in,
	   Mat &m)
{
  for (int i=0; i<5; i++)
    in >> m(i/3, i%3);
  in >> m(2, 3);

  return in;
}

void
glVertex(Vec r)
{
  glVertex2d(r(0), r(1));
}
