/** vecmat.h
 */
/*
    Smoke Vector Graphics engine
    Copyright (C) 1998-2003  Jon Harrop

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#ifndef VECMAT_H
#define VECMAT_H

#include "num.h"
#include <fstream>
#include <algorithm>

template<class T, int N>
class TinyVector
{
 private:
  T x[N];

 public:
  TinyVector() { std::fill_n(x, N, T()); }

  TinyVector(T *a) { std::copy(a, a+N, x); }

  // FIXME: We should use template partial specialization to only allow
  // n-constructors on n-vectors.

  TinyVector(T x0, T x1) { x[0]=x0; x[1]=x1; }

  TinyVector(T x0, T x1, T x2) { x[0]=x0; x[1]=x1; x[2]=x2; }

  TinyVector(T x0, T x1, T x2, T x3) { x[0]=x0; x[1]=x1; x[2]=x2; x[3]=x3; }

  const T &
    operator()(int i) const
    { /*if (i<0 || i>=N) throw;*/ return x[i]; }

  T &
    operator()(int i)
    { /*if (i<0 || i>=N) throw;*/ return x[i]; }

  TinyVector<T, N>
    operator-() const
    { T t[N]; for (int i=0; i<N; i++) t[i]=-x[i]; return TinyVector<T, N>(t); }

  TinyVector<T, N> &
    operator*=(TinyVector<T, N> r)
    { for (int i=0; i<N; i++) x[i]*=r(i); return *this; }

  TinyVector<T, N> &
    operator/=(TinyVector<T, N> r)
    { for (int i=0; i<N; i++) x[i]/=r(i); return *this; }

  TinyVector<T, N> &
    operator*=(T s)
    { for (int i=0; i<N; i++) x[i]*=s; return *this; }

  TinyVector<T, N> &
    operator/=(T s)
    { return *this*=T(1)/s; }

  TinyVector<T, N>
    operator*(TinyVector<T, N> a) const
    { return TinyVector<T, N>(*this)*=a; }

  TinyVector<T, N>
    operator/(TinyVector<T, N> a) const
    { return TinyVector<T, N>(*this)/=a; }

  TinyVector<T, N> &
    operator+=(TinyVector<T, N> a)
    { for (int i=0; i<N; i++) x[i]+=a(i); return *this; }

  TinyVector<T, N> &
    operator-=(TinyVector<T, N> a)
    { return *this+=-a; }

  TinyVector<T, N>
    operator+(TinyVector<T, N> a) const
    { return TinyVector<T, N>(*this)+=a; }

  TinyVector<T, N>
    operator-(TinyVector<T, N> a) const
    { return TinyVector<T, N>(*this)-=a; }

  bool
    operator==(TinyVector<T, N> a) const
  { for (int i=0; i<N; i++) if ((*this)(i) != a(i)) return false; return true; }

  bool
    operator!=(TinyVector<T, N> a) const
  { return !(*this==a); }
};

template<class T, int N>
  TinyVector<T,N>
  operator*(T s, TinyVector<T,N> r)
{ return TinyVector<T,N>(r)*=s; }

template<class T, int N>
  TinyVector<T,N>
  operator*(int s, TinyVector<T,N> r)
{ return TinyVector<T,N>(r)*=s; }

template<class T, int N>
  TinyVector<T,N>
  operator*(TinyVector<T,N> r, T s)
{ return TinyVector<T,N>(r)*=s; }

template<class T, int N>
  TinyVector<T,N>
  operator/(TinyVector<T,N> r, T s)
{ return TinyVector<T,N>(r)/=s; }

template<class T, int N>
  T
  dot(TinyVector<T,N> r1, TinyVector<T,N> r2)
{ T t=0; for (int i=0; i<N; i++) t+=r1(i)*r2(i); return t; }

template<class T, int N>
std::ostream &
  operator<<(std::ostream &out, TinyVector<T,N> r)
{
  for (int i=0; i<N-1; i++)
    out << r(i) << " ";
  if (N>0)
    out << r(N-1);
  return out;
}

template<class T, int N>
std::istream &
  operator>>(std::istream &in, TinyVector<T,N> &r)
{
  for (int i=0; i<N; i++)
    in >> r(i);
  return in;
}

typedef TinyVector<int, 2> iVec;

typedef TinyVector<Num, 2> Vec;

class Mat
{
 private:
  Num x[6];

 public:
  Mat() {};

  const Num &
    operator()(int i, int j) const
    {
      /*
      if (i<0 || i>=2 || j<0 || j>=3)
	throw;
      */
      return x[j + 3*i];
    }

  Num &
    operator()(int i, int j)
    {
      /*
      if (i<0 || i>=2 || j<0 || j>=3)
	throw;
      */
      return x[j + 3*i];
    }
};

/** Matrix-vector multiplication. */
Vec
operator*(const Mat a,
	  const Vec b);

/** Extract the current OpenGL model view matrix. */
void
getModelView(Mat &m);

/** Invert a 3x3 matrix in homogeneous coordinates. */
Mat
inverse(const Mat m);

/** Clockwise rotation matrix. */
Mat
rotationMatrix(const Num a);

/** Scaling matrix. */
Mat
scaleMatrix(const Num a);

/** Translation matrix. */
Mat
translationMatrix(const Vec r);

Num
length2(Vec r);

Num
length(Vec r);

/** Matrix multiplication. */
Mat
operator*(const Mat m1,
	  const Mat m2);

std::ostream &
operator<<(std::ostream &out,
	   Mat m);

void
glVertex(Vec r);

#endif /* VECMAT_H */
